@@ -57,8 +57,8 @@ public:
       return;
 
     // Check that src and dest ranks are unique and sorted
-    assert(std::is_sorted(_src.begin(), _src.end()));
-    assert(std::is_sorted(_dest.begin(), _dest.end()));
+    assert(std::ranges::is_sorted(_src));
+    assert(std::ranges::is_sorted(_dest));
 
     // Create communicators with directed edges:
     // (0) owner -> ghost,
@@ -79,17 +79,17 @@ public:
     std::span owners = map.owners();
     std::vector<std::int32_t> perm(owners.size());
     std::iota(perm.begin(), perm.end(), 0);
-    dolfinx::argsort_radix<std::int32_t>(owners, perm);
+    dolfinx::radix_sort(perm, [&owners](auto index) { return owners[index]; });
 
     // Sort (i) ghost indices and (ii) ghost index owners by rank
     // (using perm array)
     std::span ghosts = map.ghosts();
     std::vector<int> owners_sorted(owners.size());
     std::vector<std::int64_t> ghosts_sorted(owners.size());
-    std::transform(perm.begin(), perm.end(), owners_sorted.begin(),
-                   [&owners](auto idx) { return owners[idx]; });
-    std::transform(perm.begin(), perm.end(), ghosts_sorted.begin(),
-                   [&ghosts](auto idx) { return ghosts[idx]; });
+    std::ranges::transform(perm, owners_sorted.begin(),
+                           [&owners](auto idx) { return owners[idx]; });
+    std::ranges::transform(perm, ghosts_sorted.begin(),
+                           [&ghosts](auto idx) { return ghosts[idx]; });
 
     // For data associated with ghost indices, packed by owning
     // (neighbourhood) rank, compute sizes and displacements. I.e.,
@@ -139,16 +139,14 @@ public:
     const std::array<std::int64_t, 2> range = map.local_range();
 #ifndef NDEBUG
     // Check that all received indice are within the owned range
-    std::for_each(recv_buffer.begin(), recv_buffer.end(), [range](auto idx)
-                  { assert(idx >= range[0] and idx < range[1]); });
+    std::ranges::for_each(recv_buffer, [range](auto idx)
+                          { assert(idx >= range[0] and idx < range[1]); });
 #endif
 
     // Scale sizes and displacements by block size
     {
-      auto rescale = [](auto& x, int bs)
-      {
-        std::transform(x.begin(), x.end(), x.begin(),
-                       [bs](auto e) { return e *= bs; });
+      auto rescale = [](auto& x, int bs) {
+        std::ranges::transform(x, x.begin(), [bs](auto e) { return e *= bs; });
       };
       rescale(_sizes_local, bs);
       rescale(_displs_local, bs);
@@ -525,7 +523,7 @@ public:
     assert(local_buffer.size() == _local_inds.size());
     if (!_local_inds.empty())
     {
-      assert(*std::max_element(_local_inds.begin(), _local_inds.end())
+      assert(*std::ranges::max_element(_local_inds)
              < std::int32_t(local_data.size()));
     }
     scatter_rev_end(request);
