@@ -13,7 +13,6 @@
 #include <dolfinx/common/log.h>
 #include <functional>
 #include <numeric>
-#include <ufcx.h>
 #include <utility>
 #include <vector>
 
@@ -22,64 +21,11 @@ using namespace dolfinx::fem;
 
 namespace
 {
-//-----------------------------------------------------------------------------
-
-/// Check if an element is a Basix element (or a blocked element
-/// containing a Basix element)
-bool is_basix_element(const ufcx_finite_element& element)
-{
-  if (element.element_type == ufcx_basix_element)
-    return true;
-  else if (element.block_size != 1)
-  {
-    // TODO: what should happen if the element is a blocked element
-    // containing a blocked element containing a Basix element?
-    return element.sub_elements[0]->element_type == ufcx_basix_element;
-  }
-  else
-    return false;
-}
-//-----------------------------------------------------------------------------
-
-/// Check if an element is a quadrature element (or a blocked element
-/// containing a quadrature element)
-bool is_quadrature_element(const ufcx_finite_element& element)
-{
-  if (element.element_type == ufcx_quadrature_element)
-    return true;
-  else if (element.block_size != 1)
-  {
-    // TODO: what should happen if the element is a blocked element
-    // containing a blocked element containing a quadrature element?
-    return element.sub_elements[0]->element_type == ufcx_quadrature_element;
-  }
-  else
-    return false;
-}
-//-----------------------------------------------------------------------------
-
-/// Check if an element is a custom Basix element (or a blocked element
-/// containing a custom Basix element)
-bool is_basix_custom_element(const ufcx_finite_element& element)
-{
-  if (element.element_type == ufcx_basix_custom_element)
-    return true;
-  else if (element.block_size != 1)
-  {
-    // TODO: what should happen if the element is a blocked element
-    // containing a blocked element containing a Basix element?
-    return element.sub_elements[0]->element_type == ufcx_basix_custom_element;
-  }
-  else
-    return false;
-}
-//-----------------------------------------------------------------------------
-
 /// Recursively extract sub finite element
 template <std::floating_point T>
 std::shared_ptr<const FiniteElement<T>>
 _extract_sub_element(const FiniteElement<T>& finite_element,
-                     const std::vector<int>& component)
+                     std::span<const int> component)
 {
   // Check that a sub system has been specified
   if (component.empty())
@@ -115,210 +61,64 @@ _extract_sub_element(const FiniteElement<T>& finite_element,
 
   return _extract_sub_element(*sub_element, sub_component);
 }
-//-----------------------------------------------------------------------------
-
 } // namespace
 
 //-----------------------------------------------------------------------------
 template <std::floating_point T>
-FiniteElement<T>::FiniteElement(const ufcx_finite_element& e)
-    : _signature(e.signature), _space_dim(e.space_dimension),
-      _reference_value_shape(e.reference_value_shape,
-                             e.reference_value_shape + e.reference_value_rank),
-      _bs(e.block_size), _is_mixed(e.element_type == ufcx_mixed_element),
-      _symmetric(e.symmetric)
+FiniteElement<T>::FiniteElement(mesh::CellType cell_type,
+                                std::span<const geometry_type> points,
+                                std::array<std::size_t, 2> pshape,
+                                std::size_t block_size, bool symmetric)
+    : _signature("Quadrature element " + std::to_string(pshape[0]) + " "
+                 + std::to_string(block_size)),
+      _space_dim(pshape[0] * block_size), _reference_value_shape({}),
+      _bs(block_size), _is_mixed(false), _symmetric(symmetric),
+      _needs_dof_permutations(false), _needs_dof_transformations(false),
+      _entity_dofs(mesh::cell_dim(cell_type) + 1),
+      _entity_closure_dofs(mesh::cell_dim(cell_type) + 1),
+      _points(std::vector<T>(points.begin(), points.end()), pshape)
 {
-  const ufcx_shape _shape = e.cell_shape;
-  switch (_shape)
+  const int tdim = mesh::cell_dim(cell_type);
+  for (int d = 0; d <= tdim; ++d)
   {
-  case interval:
-    _cell_shape = mesh::CellType::interval;
-    break;
-  case triangle:
-    _cell_shape = mesh::CellType::triangle;
-    break;
-  case quadrilateral:
-    _cell_shape = mesh::CellType::quadrilateral;
-    break;
-  case tetrahedron:
-    _cell_shape = mesh::CellType::tetrahedron;
-    break;
-  case prism:
-    _cell_shape = mesh::CellType::prism;
-    break;
-  case hexahedron:
-    _cell_shape = mesh::CellType::hexahedron;
-    break;
-  default:
-    throw std::runtime_error(
-        "Unknown UFC cell type when building FiniteElement.");
-  }
-  assert(mesh::cell_dim(_cell_shape) == e.topological_dimension);
-
-  static const std::map<ufcx_shape, std::string> ufcx_to_cell
-      = {{vertex, "point"},         {interval, "interval"},
-         {triangle, "triangle"},    {tetrahedron, "tetrahedron"},
-         {prism, "prism"},          {quadrilateral, "quadrilateral"},
-         {hexahedron, "hexahedron"}};
-  const std::string cell_shape = ufcx_to_cell.at(e.cell_shape);
-
-  _needs_dof_transformations = false;
-  _needs_dof_permutations = false;
-  // Create all sub-elements
-  for (int i = 0; i < e.num_sub_elements; ++i)
-  {
-    ufcx_finite_element* ufcx_sub_element = e.sub_elements[i];
-    _sub_elements.push_back(
-        std::make_shared<FiniteElement<T>>(*ufcx_sub_element));
-    if (_sub_elements[i]->needs_dof_permutations())
-      _needs_dof_permutations = true;
-    if (_sub_elements[i]->needs_dof_transformations())
-      _needs_dof_transformations = true;
+    int num_entities = mesh::cell_num_entities(cell_type, d);
+    _entity_dofs[d].resize(num_entities);
+    _entity_closure_dofs[d].resize(num_entities);
   }
 
-  if (is_basix_custom_element(e))
+  for (std::size_t i = 0; i < pshape[0]; ++i)
   {
-    // Recreate the custom Basix element using information written into
-    // the generated code
-    ufcx_basix_custom_finite_element* ce = e.custom_element;
-    const basix::cell::type cell_type
-        = static_cast<basix::cell::type>(ce->cell_type);
-
-    const std::vector<std::size_t> value_shape(
-        ce->value_shape, ce->value_shape + ce->value_shape_length);
-    const std::size_t value_size = std::reduce(
-        value_shape.begin(), value_shape.end(), 1, std::multiplies{});
-
-    const int nderivs = ce->interpolation_nderivs;
-    const std::size_t nderivs_dim = basix::polyset::nderivs(cell_type, nderivs);
-
-    using array2_t = std::pair<std::vector<T>, std::array<std::size_t, 2>>;
-    using array4_t = std::pair<std::vector<T>, std::array<std::size_t, 4>>;
-    std::array<std::vector<array2_t>, 4> x;
-    std::array<std::vector<array4_t>, 4> M;
-    { // scope
-      int pt_n = 0;
-      int p_e = 0;
-      int m_e = 0;
-      const std::size_t dim = static_cast<std::size_t>(
-          basix::cell::topological_dimension(cell_type));
-      for (std::size_t d = 0; d <= dim; ++d)
-      {
-        const int num_entities = basix::cell::num_sub_entities(cell_type, d);
-        for (int entity = 0; entity < num_entities; ++entity)
-        {
-          std::size_t npts = ce->npts[pt_n + entity];
-          std::size_t ndofs = ce->ndofs[pt_n + entity];
-
-          std::array pshape = {npts, dim};
-          auto& pts
-              = x[d].emplace_back(std::vector<T>(pshape[0] * pshape[1]), pshape)
-                    .first;
-          std::copy_n(ce->x + p_e, pts.size(), pts.begin());
-          p_e += pts.size();
-
-          std::array mshape = {ndofs, value_size, npts, nderivs_dim};
-          std::size_t msize
-              = std::reduce(mshape.begin(), mshape.end(), 1, std::multiplies{});
-          auto& mat = M[d].emplace_back(std::vector<T>(msize), mshape).first;
-          std::copy_n(ce->M + m_e, mat.size(), mat.begin());
-          m_e += mat.size();
-        }
-
-        pt_n += num_entities;
-      }
-    }
-
-    using cmdspan2_t = MDSPAN_IMPL_STANDARD_NAMESPACE::mdspan<
-        const T, MDSPAN_IMPL_STANDARD_NAMESPACE::dextents<std::size_t, 2>>;
-    using cmdspan4_t = MDSPAN_IMPL_STANDARD_NAMESPACE::mdspan<
-        const T, MDSPAN_IMPL_STANDARD_NAMESPACE::dextents<std::size_t, 4>>;
-
-    std::array<std::vector<cmdspan2_t>, 4> _x;
-    for (std::size_t i = 0; i < x.size(); ++i)
-      for (auto& [buffer, shape] : x[i])
-        _x[i].push_back(cmdspan2_t(buffer.data(), shape));
-
-    std::array<std::vector<cmdspan4_t>, 4> _M;
-    for (std::size_t i = 0; i < M.size(); ++i)
-      for (auto& [buffer, shape] : M[i])
-        _M[i].push_back(cmdspan4_t(buffer.data(), shape));
-
-    std::vector<T> wcoeffs_b(ce->wcoeffs_rows * ce->wcoeffs_cols);
-    cmdspan2_t wcoeffs(wcoeffs_b.data(), ce->wcoeffs_rows, ce->wcoeffs_cols);
-    std::copy_n(ce->wcoeffs, wcoeffs_b.size(), wcoeffs_b.begin());
-
-    _element = std::make_unique<basix::FiniteElement<T>>(
-        basix::create_custom_element(
-            cell_type, value_shape, wcoeffs, _x, _M, nderivs,
-            static_cast<basix::maps::type>(ce->map_type),
-            static_cast<basix::sobolev::space>(ce->sobolev_space),
-            ce->discontinuous, ce->embedded_subdegree, ce->embedded_superdegree,
-            static_cast<basix::polyset::type>(ce->polyset_type)));
-    _needs_dof_transformations
-        = !_element->dof_transformations_are_identity()
-          and !_element->dof_transformations_are_permutations();
-
-    _needs_dof_permutations
-        = !_element->dof_transformations_are_identity()
-          and _element->dof_transformations_are_permutations();
-  }
-  else if (is_basix_element(e))
-  {
-    _element
-        = std::make_unique<basix::FiniteElement<T>>(basix::create_element<T>(
-            static_cast<basix::element::family>(e.basix_family),
-            static_cast<basix::cell::type>(e.basix_cell), e.degree,
-            static_cast<basix::element::lagrange_variant>(e.lagrange_variant),
-            static_cast<basix::element::dpc_variant>(e.dpc_variant),
-            e.discontinuous));
-    _needs_dof_transformations
-        = !_element->dof_transformations_are_identity()
-          and !_element->dof_transformations_are_permutations();
-    _needs_dof_permutations
-        = !_element->dof_transformations_are_identity()
-          and _element->dof_transformations_are_permutations();
-  }
-
-  if (is_quadrature_element(e))
-  {
-    assert(e.custom_quadrature);
-    ufcx_quadrature_rule* qr = e.custom_quadrature;
-    std::size_t npts = qr->npts;
-    std::size_t tdim = qr->topological_dimension;
-    std::array qshape = {npts, tdim};
-    _points = std::make_pair(std::vector<T>(qshape[0] * qshape[1]), qshape);
-    std::copy_n(qr->points, _points.first.size(), _points.first.begin());
+    _entity_dofs[tdim][0].push_back(i);
+    _entity_closure_dofs[tdim][0].push_back(i);
   }
 }
 //-----------------------------------------------------------------------------
 template <std::floating_point T>
 FiniteElement<T>::FiniteElement(const basix::FiniteElement<T>& element,
-                                const std::size_t block_size,
-                                const bool symmetric)
-    : _reference_value_shape(element.value_shape()), _bs(block_size),
-      _is_mixed(false), _symmetric(symmetric)
+                                std::size_t block_size, bool symmetric)
+    : _space_dim(block_size * element.dim()),
+      _reference_value_shape(element.value_shape()), _bs(block_size),
+      _is_mixed(false),
+      _element(std::make_unique<basix::FiniteElement<T>>(element)),
+      _symmetric(symmetric),
+      _needs_dof_permutations(
+          !_element->dof_transformations_are_identity()
+          and _element->dof_transformations_are_permutations()),
+      _needs_dof_transformations(
+          !_element->dof_transformations_are_identity()
+          and !_element->dof_transformations_are_permutations()),
+      _entity_dofs(element.entity_dofs()),
+      _entity_closure_dofs(element.entity_closure_dofs())
 {
-  _space_dim = _bs * element.dim();
-
   // Create all sub-elements
   if (_bs > 1)
   {
-    for (int i = 0; i < _bs; ++i)
-    {
-      _sub_elements.push_back(std::make_shared<FiniteElement<T>>(element, 1));
-    }
+    _sub_elements
+        = std::vector<std::shared_ptr<const FiniteElement<geometry_type>>>(
+            _bs, std::make_shared<FiniteElement<T>>(element, 1));
     _reference_value_shape = {block_size};
   }
 
-  _element = std::make_unique<basix::FiniteElement<T>>(element);
-  assert(_element);
-  _needs_dof_transformations
-      = !_element->dof_transformations_are_identity()
-        and !_element->dof_transformations_are_permutations();
-  _needs_dof_permutations
-      = !_element->dof_transformations_are_identity()
-        and _element->dof_transformations_are_permutations();
   std::string family;
   switch (_element->family())
   {
@@ -337,6 +137,65 @@ FiniteElement<T>::FiniteElement(const basix::FiniteElement<T>& element,
 }
 //-----------------------------------------------------------------------------
 template <std::floating_point T>
+FiniteElement<T>::FiniteElement(
+    const std::vector<std::shared_ptr<const FiniteElement<T>>>& elements)
+    : _space_dim(0), _sub_elements(elements), _bs(1), _is_mixed(true),
+      _symmetric(false), _needs_dof_permutations(false),
+      _needs_dof_transformations(false)
+{
+  std::size_t vsize = 0;
+  _signature = "Mixed element (";
+
+  const std::vector<std::vector<std::vector<int>>>& ed
+      = elements[0]->entity_dofs();
+  _entity_dofs.resize(ed.size());
+  _entity_closure_dofs.resize(ed.size());
+  for (std::size_t i = 0; i < ed.size(); ++i)
+  {
+    _entity_dofs[i].resize(ed[i].size());
+    _entity_closure_dofs[i].resize(ed[i].size());
+  }
+
+  int dof_offset = 0;
+  for (auto& e : elements)
+  {
+    vsize += e->reference_value_size();
+    _signature += e->signature() + ", ";
+
+    if (e->needs_dof_permutations())
+      _needs_dof_permutations = true;
+    if (e->needs_dof_transformations())
+      _needs_dof_transformations = true;
+
+    const std::size_t sub_bs = e->block_size();
+    for (std::size_t i = 0; i < _entity_dofs.size(); ++i)
+    {
+      for (std::size_t j = 0; j < _entity_dofs[i].size(); ++j)
+      {
+        const std::vector<int> sub_ed = e->entity_dofs()[i][j];
+        const std::vector<int> sub_ecd = e->entity_closure_dofs()[i][j];
+        for (auto k : sub_ed)
+        {
+          for (std::size_t b = 0; b < sub_bs; ++b)
+            _entity_dofs[i][j].push_back(dof_offset + k * sub_bs + b);
+        }
+        for (auto k : sub_ecd)
+        {
+          for (std::size_t b = 0; b < sub_bs; ++b)
+            _entity_closure_dofs[i][j].push_back(dof_offset + k * sub_bs + b);
+        }
+      }
+    }
+
+    dof_offset += e->space_dimension();
+  }
+
+  _space_dim = dof_offset;
+  _reference_value_shape = {vsize};
+  _signature += ")";
+}
+//-----------------------------------------------------------------------------
+template <std::floating_point T>
 bool FiniteElement<T>::operator==(const FiniteElement& e) const
 {
   if (!_element or !e._element)
@@ -361,24 +220,33 @@ std::string FiniteElement<T>::signature() const noexcept
 }
 //-----------------------------------------------------------------------------
 template <std::floating_point T>
-mesh::CellType FiniteElement<T>::cell_shape() const noexcept
-{
-  return _cell_shape;
-}
-//-----------------------------------------------------------------------------
-template <std::floating_point T>
 int FiniteElement<T>::space_dimension() const noexcept
 {
   return _space_dim;
 }
 //-----------------------------------------------------------------------------
 template <std::floating_point T>
-std::span<const std::size_t> FiniteElement<T>::reference_value_shape() const
+std::span<const std::size_t>
+FiniteElement<T>::reference_value_shape() const noexcept
 {
   return _reference_value_shape;
 }
 //-----------------------------------------------------------------------------
 template <std::floating_point T>
+const std::vector<std::vector<std::vector<int>>>&
+FiniteElement<T>::entity_dofs() const noexcept
+{
+  return _entity_dofs;
+}
+//-----------------------------------------------------------------------------
+template <std::floating_point T>
+const std::vector<std::vector<std::vector<int>>>&
+FiniteElement<T>::entity_closure_dofs() const noexcept
+{
+  return _entity_closure_dofs;
+}
+//-----------------------------------------------------------------------------
+template <std::floating_point T>
 bool FiniteElement<T>::symmetric() const
 {
   return _symmetric;
@@ -440,8 +308,8 @@ FiniteElement<T>::extract_sub_element(const std::vector<int>& component) const
 {
   // Recursively extract sub element
   auto sub_finite_element = _extract_sub_element(*this, component);
-  DLOG(INFO) << "Extracted finite element for sub-system: "
-             << sub_finite_element->signature().c_str();
+  spdlog::debug("Extracted finite element for sub-system: {}",
+                sub_finite_element->signature().c_str());
   return sub_finite_element;
 }
 //-----------------------------------------------------------------------------
