@@ -15,6 +15,7 @@
 #include "Function.h"
 #include "FunctionSpace.h"
 #include "sparsitybuild.h"
+#include <algorithm>
 #include <array>
 #include <concepts>
 #include <dolfinx/common/types.h>
@@ -82,36 +83,32 @@ get_cell_facet_pairs(std::int32_t f, std::span<const std::int32_t> cells,
 
 } // namespace impl
 
-/// @brief Given an integral type and mesh tag data, compute the
+/// @brief Given an integral type and a set of entities, compute the
 /// entities that should be integrated over.
 ///
-/// This function returns as list `[(id, entities)]`, where `entities`
-/// are the entities in `meshtags` with tag value `id`. For cell
+/// This function returns a list `[(id, entities)]`. For cell
 /// integrals `entities` are the cell indices. For exterior facet
 /// integrals, `entities` is a list of `(cell_index, local_facet_index)`
 /// pairs. For interior facet integrals, `entities` is a list of
-/// `(cell_index0, local_facet_index0, cell_index1,
-/// local_facet_index1)`.
+/// `(cell_index0, local_facet_index0, cell_index1, local_facet_index1)`.
+/// `id` refers to the subdomain id used in the definition of the integration
+/// measures of the variational form.
 ///
 /// @note Owned mesh entities only are returned. Ghost entities are not
 /// included.
 ///
 /// @param[in] integral_type Integral type
 /// @param[in] topology Mesh topology
-/// @param[in] entities List of tagged mesh entities
-/// @param[in] dim Topological dimension of tagged entities
-/// @param[in] values Value associated with each entity
-/// @return List of `(integral id, entities)` pairs
-/// @pre The topological dimension of the integral entity type and the
-/// topological dimension of mesh tag data must be equal.
+/// @param[in] entities List of mesh entities
+/// @param[in] dim Topological dimension of entities
+/// @return List of integration entities
 /// @pre For facet integrals, the topology facet-to-cell and
 /// cell-to-facet connectivity must be computed before calling this
 /// function.
-std::vector<std::pair<int, std::vector<std::int32_t>>>
+std::vector<std::int32_t>
 compute_integration_domains(IntegralType integral_type,
                             const mesh::Topology& topology,
-                            std::span<const std::int32_t> entities, int dim,
-                            std::span<const int> values);
+                            std::span<const std::int32_t> entities, int dim);
 
 /// @brief Extract test (0) and trial (1) function spaces pairs for each
 /// bilinear form for a rectangular array of forms.
@@ -239,11 +236,40 @@ la::SparsityPattern create_sparsity_pattern(const Form<T, U>& a)
   return pattern;
 }
 
-/// Create an ElementDofLayout from a ufcx_dofmap
-ElementDofLayout create_element_dof_layout(const ufcx_dofmap& dofmap,
-                                           const mesh::CellType cell_type,
+/// Create an ElementDofLayout from a FiniteElement
+template <std::floating_point T>
+ElementDofLayout create_element_dof_layout(const fem::FiniteElement<T>& element,
                                            const std::vector<int>& parent_map
-                                           = {});
+                                           = {})
+{
+  // Create subdofmaps and compute offset
+  std::vector<int> offsets(1, 0);
+  std::vector<dolfinx::fem::ElementDofLayout> sub_doflayout;
+  int bs = element.block_size();
+  for (int i = 0; i < element.num_sub_elements(); ++i)
+  {
+    // The ith sub-element. For mixed elements this is subelements()[i]. For
+    // blocked elements, the sub-element will always be the same, so we'll use
+    // sub_elements()[0]
+    std::shared_ptr<const fem::FiniteElement<T>> sub_e
+        = element.sub_elements()[bs > 1 ? 0 : i];
+
+    // In a mixed element DOFs are ordered element by element, so the offset to
+    // the next sub-element is sub_e->space_dimension(). Blocked elements use
+    // xxyyzz ordering, so the offset to the next sub-element is 1
+
+    std::vector<int> parent_map_sub(sub_e->space_dimension(), offsets.back());
+    for (std::size_t j = 0; j < parent_map_sub.size(); ++j)
+      parent_map_sub[j] += bs * j;
+    offsets.push_back(offsets.back() + (bs > 1 ? 1 : sub_e->space_dimension()));
+    sub_doflayout.push_back(
+        dolfinx::fem::create_element_dof_layout(*sub_e, parent_map_sub));
+  }
+
+  return ElementDofLayout(bs, element.entity_dofs(),
+                          element.entity_closure_dofs(), parent_map,
+                          sub_doflayout);
+}
 
 /// @brief Create a dof map on mesh
 /// @param[in] comm MPI communicator
@@ -331,20 +357,17 @@ Form<T, U> create_form_factory(
   }
 
   // Check argument function spaces
-#ifndef NDEBUG
   for (std::size_t i = 0; i < spaces.size(); ++i)
   {
     assert(spaces[i]->element());
-    ufcx_finite_element* ufcx_element = ufcx_form.finite_elements[i];
-    assert(ufcx_element);
-    if (std::string(ufcx_element->signature)
-        != spaces[i]->element()->signature())
+    if (auto element_hash = ufcx_form.finite_element_hashes[i];
+        element_hash != 0
+        and element_hash != spaces[i]->element()->basix_element().hash())
     {
-      throw std::runtime_error(
-          "Cannot create form. Wrong type of function space for argument.");
+      throw std::runtime_error("Cannot create form. Elements are different to "
+                               "those used to compile the form.");
     }
   }
-#endif
 
   // Extract mesh from FunctionSpace, and check they are the same
   if (!mesh and !spaces.empty())
@@ -398,9 +421,18 @@ Form<T, U> create_form_factory(
           = ufcx_form.form_integrals[integral_offsets[cell] + i];
       assert(integral);
 
+      // Build list of active coefficients
+      std::vector<int> active_coeffs;
+      for (int j = 0; j < ufcx_form.num_coefficients; ++j)
+      {
+        if (integral->enabled_coefficients[j])
+          active_coeffs.push_back(j);
+      }
+
       kern_t k = nullptr;
       if constexpr (std::is_same_v<T, float>)
         k = integral->tabulate_tensor_float32;
+#ifndef DOLFINX_NO_STDC_COMPLEX_KERNELS
       else if constexpr (std::is_same_v<T, std::complex<float>>)
       {
         k = reinterpret_cast<void (*)(
@@ -408,8 +440,10 @@ Form<T, U> create_form_factory(
             const typename scalar_value_type<T>::value_type*, const int*,
             const unsigned char*)>(integral->tabulate_tensor_complex64);
       }
+#endif // DOLFINX_NO_STDC_COMPLEX_KERNELS
       else if constexpr (std::is_same_v<T, double>)
         k = integral->tabulate_tensor_float64;
+#ifndef DOLFINX_NO_STDC_COMPLEX_KERNELS
       else if constexpr (std::is_same_v<T, std::complex<double>>)
       {
         k = reinterpret_cast<void (*)(
@@ -417,6 +451,8 @@ Form<T, U> create_form_factory(
             const typename scalar_value_type<T>::value_type*, const int*,
             const unsigned char*)>(integral->tabulate_tensor_complex128);
       }
+#endif // DOLFINX_NO_STDC_COMPLEX_KERNELS
+
       if (!k)
       {
         throw std::runtime_error(
@@ -430,16 +466,16 @@ Form<T, U> create_form_factory(
         assert(topology->index_map(tdim));
         default_cells.resize(topology->index_map(tdim)->size_local(), 0);
         std::iota(default_cells.begin(), default_cells.end(), 0);
-        itg.first->second.emplace_back(id, k, default_cells);
+        itg.first->second.emplace_back(id, k, default_cells, active_coeffs);
       }
       else if (sd != subdomains.end())
       {
         // NOTE: This requires that pairs are sorted
-        auto it = std::lower_bound(sd->second.begin(), sd->second.end(), id,
-                                   [](auto& pair, auto val)
-                                   { return pair.first < val; });
+        auto it
+            = std::ranges::lower_bound(sd->second, id, std::less<>{},
+                                       [](const auto& a) { return a.first; });
         if (it != sd->second.end() and it->first == id)
-          itg.first->second.emplace_back(id, k, it->second);
+          itg.first->second.emplace_back(id, k, it->second, active_coeffs);
       }
 
       if (integral->needs_facet_permutations)
@@ -461,10 +497,17 @@ Form<T, U> create_form_factory(
       ufcx_integral* integral
           = ufcx_form.form_integrals[integral_offsets[exterior_facet] + i];
       assert(integral);
+      std::vector<int> active_coeffs;
+      for (int j = 0; j < ufcx_form.num_coefficients; ++j)
+      {
+        if (integral->enabled_coefficients[j])
+          active_coeffs.push_back(j);
+      }
 
       kern_t k = nullptr;
       if constexpr (std::is_same_v<T, float>)
         k = integral->tabulate_tensor_float32;
+#ifndef DOLFINX_NO_STDC_COMPLEX_KERNELS
       else if constexpr (std::is_same_v<T, std::complex<float>>)
       {
         k = reinterpret_cast<void (*)(
@@ -472,8 +515,10 @@ Form<T, U> create_form_factory(
             const typename scalar_value_type<T>::value_type*, const int*,
             const unsigned char*)>(integral->tabulate_tensor_complex64);
       }
+#endif // DOLFINX_NO_STDC_COMPLEX_KERNELS
       else if constexpr (std::is_same_v<T, double>)
         k = integral->tabulate_tensor_float64;
+#ifndef DOLFINX_NO_STDC_COMPLEX_KERNELS
       else if constexpr (std::is_same_v<T, std::complex<double>>)
       {
         k = reinterpret_cast<void (*)(
@@ -481,6 +526,7 @@ Form<T, U> create_form_factory(
             const typename scalar_value_type<T>::value_type*, const int*,
             const unsigned char*)>(integral->tabulate_tensor_complex128);
       }
+#endif // DOLFINX_NO_STDC_COMPLEX_KERNELS
       assert(k);
 
       // Build list of entities to assembler over
@@ -501,16 +547,17 @@ Form<T, U> create_form_factory(
           default_facets_ext.insert(default_facets_ext.end(), pair.begin(),
                                     pair.end());
         }
-        itg.first->second.emplace_back(id, k, default_facets_ext);
+        itg.first->second.emplace_back(id, k, default_facets_ext,
+                                       active_coeffs);
       }
       else if (sd != subdomains.end())
       {
         // NOTE: This requires that pairs are sorted
-        auto it = std::lower_bound(sd->second.begin(), sd->second.end(), id,
-                                   [](auto& pair, auto val)
-                                   { return pair.first < val; });
+        auto it
+            = std::ranges::lower_bound(sd->second, id, std::less<>{},
+                                       [](const auto& a) { return a.first; });
         if (it != sd->second.end() and it->first == id)
-          itg.first->second.emplace_back(id, k, it->second);
+          itg.first->second.emplace_back(id, k, it->second, active_coeffs);
       }
 
       if (integral->needs_facet_permutations)
@@ -526,16 +573,38 @@ Form<T, U> create_form_factory(
                              num_integrals_type[interior_facet]);
     auto itg = integrals.insert({IntegralType::interior_facet, {}});
     auto sd = subdomains.find(IntegralType::interior_facet);
+
+    // Create indicator for interprocess facets
+    std::vector<std::int8_t> interprocess_marker;
+    if (num_integrals_type[interior_facet] > 0)
+    {
+      assert(topology->index_map(tdim - 1));
+      const std::vector<std::int32_t>& interprocess_facets
+          = topology->interprocess_facets();
+      std::int32_t num_facets = topology->index_map(tdim - 1)->size_local()
+                                + topology->index_map(tdim - 1)->num_ghosts();
+      interprocess_marker.resize(num_facets, 0);
+      std::ranges::for_each(interprocess_facets, [&interprocess_marker](auto f)
+                            { interprocess_marker[f] = 1; });
+    }
+
     for (int i = 0; i < num_integrals_type[interior_facet]; ++i)
     {
       const int id = ids[i];
       ufcx_integral* integral
           = ufcx_form.form_integrals[integral_offsets[interior_facet] + i];
       assert(integral);
+      std::vector<int> active_coeffs;
+      for (int j = 0; j < ufcx_form.num_coefficients; ++j)
+      {
+        if (integral->enabled_coefficients[j])
+          active_coeffs.push_back(j);
+      }
 
       kern_t k = nullptr;
       if constexpr (std::is_same_v<T, float>)
         k = integral->tabulate_tensor_float32;
+#ifndef DOLFINX_NO_STDC_COMPLEX_KERNELS
       else if constexpr (std::is_same_v<T, std::complex<float>>)
       {
         k = reinterpret_cast<void (*)(
@@ -543,8 +612,10 @@ Form<T, U> create_form_factory(
             const typename scalar_value_type<T>::value_type*, const int*,
             const unsigned char*)>(integral->tabulate_tensor_complex64);
       }
+#endif // DOLFINX_NO_STDC_COMPLEX_KERNELS
       else if constexpr (std::is_same_v<T, double>)
         k = integral->tabulate_tensor_float64;
+#ifndef DOLFINX_NO_STDC_COMPLEX_KERNELS
       else if constexpr (std::is_same_v<T, std::complex<double>>)
       {
         k = reinterpret_cast<void (*)(
@@ -552,6 +623,7 @@ Form<T, U> create_form_factory(
             const typename scalar_value_type<T>::value_type*, const int*,
             const unsigned char*)>(integral->tabulate_tensor_complex128);
       }
+#endif // DOLFINX_NO_STDC_COMPLEX_KERNELS
       assert(k);
 
       // Build list of entities to assembler over
@@ -574,16 +646,24 @@ Form<T, U> create_form_factory(
             default_facets_int.insert(default_facets_int.end(), pairs.begin(),
                                       pairs.end());
           }
+          else if (interprocess_marker[f])
+          {
+            throw std::runtime_error(
+                "Cannot compute interior facet integral over interprocess "
+                "facet. Please use ghost mode shared facet when creating the "
+                "mesh");
+          }
         }
-        itg.first->second.emplace_back(id, k, default_facets_int);
+        itg.first->second.emplace_back(id, k, default_facets_int,
+                                       active_coeffs);
       }
       else if (sd != subdomains.end())
       {
-        auto it = std::lower_bound(sd->second.begin(), sd->second.end(), id,
-                                   [](auto& pair, auto val)
-                                   { return pair.first < val; });
+        auto it
+            = std::ranges::lower_bound(sd->second, id, std::less<>{},
+                                       [](const auto& a) { return a.first; });
         if (it != sd->second.end() and it->first == id)
-          itg.first->second.emplace_back(id, k, it->second);
+          itg.first->second.emplace_back(id, k, it->second, active_coeffs);
       }
 
       if (integral->needs_facet_permutations)
@@ -614,6 +694,8 @@ Form<T, U> create_form_factory(
 /// @param[in] constants Spatial constants in the form (by name).
 /// @param[in] subdomains Subdomain markers.
 /// @pre Each value in `subdomains` must be sorted by domain id.
+/// @param[in] entity_maps The entity maps for the form. Empty for
+/// single domain problems.
 /// @param[in] mesh Mesh of the domain. This is required if the form has
 /// no arguments, e.g. a functional.
 /// @return A Form
@@ -628,6 +710,8 @@ Form<T, U> create_form(
         IntegralType,
         std::vector<std::pair<std::int32_t, std::span<const std::int32_t>>>>&
         subdomains,
+    const std::map<std::shared_ptr<const mesh::Mesh<U>>,
+                   std::span<const std::int32_t>>& entity_maps,
     std::shared_ptr<const mesh::Mesh<U>> mesh = nullptr)
 {
   // Place coefficients in appropriate order
@@ -644,6 +728,7 @@ Form<T, U> create_form(
   }
 
   // Place constants in appropriate order
+
   std::vector<std::shared_ptr<const Constant<T>>> const_map;
   for (const std::string& name : get_constant_names(ufcx_form))
   {
@@ -654,7 +739,7 @@ Form<T, U> create_form(
   }
 
   return create_form_factory(ufcx_form, spaces, coeff_map, const_map,
-                             subdomains, {}, mesh);
+                             subdomains, entity_maps, mesh);
 }
 
 /// @brief Create a Form using a factory function that returns a pointer
@@ -669,6 +754,8 @@ Form<T, U> create_form(
 /// @param[in] constants Spatial constants in the form (by name),
 /// @param[in] subdomains Subdomain markers.
 /// @pre Each value in `subdomains` must be sorted by domain id.
+/// @param[in] entity_maps The entity maps for the form. Empty for
+/// single domain problems.
 /// @param[in] mesh Mesh of the domain. This is required if the form has
 /// no arguments, e.g. a functional.
 /// @return A Form
@@ -683,11 +770,13 @@ Form<T, U> create_form(
         IntegralType,
         std::vector<std::pair<std::int32_t, std::span<const std::int32_t>>>>&
         subdomains,
+    const std::map<std::shared_ptr<const mesh::Mesh<U>>,
+                   std::span<const std::int32_t>>& entity_maps,
     std::shared_ptr<const mesh::Mesh<U>> mesh = nullptr)
 {
   ufcx_form* form = fptr();
   Form<T, U> L = create_form<T, U>(*form, spaces, coefficients, constants,
-                                   subdomains, mesh);
+                                   subdomains, entity_maps, mesh);
   std::free(form);
   return L;
 }
@@ -717,6 +806,10 @@ FunctionSpace<T> create_functionspace(
         "Cannot specify value shape for non-scalar base element.");
   }
 
+  if (mesh::cell_type_from_basix_type(e.cell_type())
+      != mesh->topology()->cell_type())
+    throw std::runtime_error("Cell type of element and mesh must match.");
+
   std::size_t bs = value_shape.empty()
                        ? 1
                        : std::accumulate(value_shape.begin(), value_shape.end(),
@@ -760,68 +853,6 @@ FunctionSpace<T> create_functionspace(
   return FunctionSpace(mesh, _e, dofmap, _value_shape);
 }
 
-/// @brief Create a FunctionSpace from UFC data.
-/// @param[in] fptr Pointer to a ufcx_function_space_create function.
-/// @param[in] function_name Name of a function whose function space is to
-/// create. Function name is the name of the Python variable for
-/// ufl.Coefficient, ufl.TrialFunction or ufl.TestFunction as defined in
-/// the UFL file.
-/// @param[in] mesh Mesh
-/// @param[in] reorder_fn Graph reordering function to call on the
-/// dofmap. If `nullptr`, the default re-ordering is used.
-/// @return The created function space.
-template <std::floating_point T>
-FunctionSpace<T> create_functionspace(
-    ufcx_function_space* (*fptr)(const char*), const std::string& function_name,
-    std::shared_ptr<mesh::Mesh<T>> mesh,
-    std::function<std::vector<int>(const graph::AdjacencyList<std::int32_t>&)>
-        reorder_fn
-    = nullptr)
-{
-  ufcx_function_space* space = fptr(function_name.c_str());
-  if (!space)
-  {
-    throw std::runtime_error(
-        "Could not create UFC function space with function name "
-        + function_name);
-  }
-
-  ufcx_finite_element* ufcx_element = space->finite_element;
-  assert(ufcx_element);
-  std::vector<std::size_t> value_shape(space->value_shape,
-                                       space->value_shape + space->value_rank);
-
-  const auto& geometry = mesh->geometry();
-  auto& cmap = geometry.cmap();
-  if (space->geometry_degree != cmap.degree()
-      or static_cast<basix::cell::type>(space->geometry_basix_cell)
-             != mesh::cell_type_to_basix_type(cmap.cell_shape())
-      or static_cast<basix::element::lagrange_variant>(
-             space->geometry_basix_variant)
-             != cmap.variant())
-  {
-    throw std::runtime_error("UFL mesh and CoordinateElement do not match.");
-  }
-
-  auto element = std::make_shared<FiniteElement<T>>(*ufcx_element);
-  assert(element);
-  ufcx_dofmap* ufcx_map = space->dofmap;
-  assert(ufcx_map);
-  const auto topology = mesh->topology();
-  assert(topology);
-  ElementDofLayout layout
-      = create_element_dof_layout(*ufcx_map, topology->cell_type());
-
-  std::function<void(std::span<std::int32_t>, std::uint32_t)> permute_inv;
-  if (element->needs_dof_permutations())
-    permute_inv = element->dof_permutation_fn(true, true);
-  return FunctionSpace(
-      mesh, element,
-      std::make_shared<DofMap>(create_dofmap(mesh->comm(), layout, *topology,
-                                             permute_inv, reorder_fn)),
-      value_shape);
-}
-
 /// @private
 namespace impl
 {
@@ -843,7 +874,7 @@ get_cell_orientation_info(const Function<T, U>& coefficient)
 }
 
 /// Pack a single coefficient for a single cell
-template <dolfinx::scalar T, int _bs>
+template <int _bs, dolfinx::scalar T>
 void pack(std::span<T> coeffs, std::int32_t cell, int bs, std::span<const T> v,
           std::span<const std::uint32_t> cell_info, const DofMap& dofmap,
           auto transform)
@@ -860,6 +891,7 @@ void pack(std::span<T> coeffs, std::int32_t cell, int bs, std::span<const T> v,
     }
     else
     {
+      assert(_bs == bs);
       const int pos_c = _bs * i;
       const int pos_v = _bs * dofs[i];
       for (int k = 0; k < _bs; ++k)
@@ -917,36 +949,47 @@ void pack_coefficient_entity(std::span<T> c, int cstride,
     for (std::size_t e = 0; e < entities.size(); e += estride)
     {
       auto entity = entities.subspan(e, estride);
-      std::int32_t cell = fetch_cells(entity);
-      auto cell_coeff = c.subspan((e / estride) * cstride + offset, space_dim);
-      pack<T, 1>(cell_coeff, cell, bs, v, cell_info, dofmap, transformation);
+      if (std::int32_t cell = fetch_cells(entity); cell >= 0)
+      {
+        auto cell_coeff
+            = c.subspan((e / estride) * cstride + offset, space_dim);
+        pack<1>(cell_coeff, cell, bs, v, cell_info, dofmap, transformation);
+      }
     }
     break;
   case 2:
     for (std::size_t e = 0; e < entities.size(); e += estride)
     {
       auto entity = entities.subspan(e, estride);
-      std::int32_t cell = fetch_cells(entity);
-      auto cell_coeff = c.subspan((e / estride) * cstride + offset, space_dim);
-      pack<T, 2>(cell_coeff, cell, bs, v, cell_info, dofmap, transformation);
+      if (std::int32_t cell = fetch_cells(entity); cell >= 0)
+      {
+        auto cell_coeff
+            = c.subspan((e / estride) * cstride + offset, space_dim);
+        pack<2>(cell_coeff, cell, bs, v, cell_info, dofmap, transformation);
+      }
     }
     break;
   case 3:
     for (std::size_t e = 0; e < entities.size(); e += estride)
     {
       auto entity = entities.subspan(e, estride);
-      std::int32_t cell = fetch_cells(entity);
-      auto cell_coeff = c.subspan(e / estride * cstride + offset, space_dim);
-      pack<T, 3>(cell_coeff, cell, bs, v, cell_info, dofmap, transformation);
+      if (std::int32_t cell = fetch_cells(entity); cell >= 0)
+      {
+        auto cell_coeff = c.subspan(e / estride * cstride + offset, space_dim);
+        pack<3>(cell_coeff, cell, bs, v, cell_info, dofmap, transformation);
+      }
     }
     break;
   default:
     for (std::size_t e = 0; e < entities.size(); e += estride)
     {
       auto entity = entities.subspan(e, estride);
-      std::int32_t cell = fetch_cells(entity);
-      auto cell_coeff = c.subspan((e / estride) * cstride + offset, space_dim);
-      pack<T, -1>(cell_coeff, cell, bs, v, cell_info, dofmap, transformation);
+      if (std::int32_t cell = fetch_cells(entity); cell >= 0)
+      {
+        auto cell_coeff
+            = c.subspan((e / estride) * cstride + offset, space_dim);
+        pack<-1>(cell_coeff, cell, bs, v, cell_info, dofmap, transformation);
+      }
     }
     break;
   }
@@ -1013,7 +1056,7 @@ allocate_coefficient_storage(const Form<T, U>& form)
 /// @param[in] form The Form
 /// @param[in] integral_type Type of integral
 /// @param[in] id The id of the integration domain
-/// @param[in] c The coefficient array
+/// @param[in,out] c The coefficient array
 /// @param[in] cstride The coefficient stride
 template <dolfinx::scalar T, std::floating_point U>
 void pack_coefficients(const Form<T, U>& form, IntegralType integral_type,
@@ -1024,17 +1067,43 @@ void pack_coefficients(const Form<T, U>& form, IntegralType integral_type,
       = form.coefficients();
   const std::vector<int> offsets = form.coefficient_offsets();
 
+  // Indicator for packing coefficients
+  std::vector<int> active_coefficient(coefficients.size(), 0);
   if (!coefficients.empty())
   {
     switch (integral_type)
     {
     case IntegralType::cell:
     {
+      // Get indicator for all coefficients that are active in cell
+      // integrals
+      for (std::size_t i = 0; i < form.num_integrals(IntegralType::cell); ++i)
+      {
+        for (auto idx : form.active_coeffs(IntegralType::cell, i))
+          active_coefficient[idx] = 1;
+      }
+
       // Iterate over coefficients
       for (std::size_t coeff = 0; coeff < coefficients.size(); ++coeff)
       {
+        if (!active_coefficient[coeff])
+          continue;
+
+        // Get coefficient mesh
         auto mesh = coefficients[coeff]->function_space()->mesh();
         assert(mesh);
+
+        // Other integrals in the form might have coefficients defined over
+        // entities of codim > 0, which don't make sense for cell integrals, so
+        // don't pack them.
+        if (const int codim
+            = form.mesh()->topology()->dim() - mesh->topology()->dim();
+            codim > 0)
+        {
+          throw std::runtime_error("Should not be packing coefficients with "
+                                   "codim>0 in a cell integral");
+        }
+
         std::vector<std::int32_t> cells
             = form.domain(IntegralType::cell, id, *mesh);
         std::span<const std::uint32_t> cell_info
@@ -1047,9 +1116,21 @@ void pack_coefficients(const Form<T, U>& form, IntegralType integral_type,
     }
     case IntegralType::exterior_facet:
     {
+      // Get indicator for all coefficients that are active in exterior
+      // facet integrals
+      for (std::size_t i = 0;
+           i < form.num_integrals(IntegralType::exterior_facet); ++i)
+      {
+        for (auto idx : form.active_coeffs(IntegralType::exterior_facet, i))
+          active_coefficient[idx] = 1;
+      }
+
       // Iterate over coefficients
       for (std::size_t coeff = 0; coeff < coefficients.size(); ++coeff)
       {
+        if (!active_coefficient[coeff])
+          continue;
+
         auto mesh = coefficients[coeff]->function_space()->mesh();
         std::vector<std::int32_t> facets
             = form.domain(IntegralType::exterior_facet, id, *mesh);
@@ -1063,12 +1144,25 @@ void pack_coefficients(const Form<T, U>& form, IntegralType integral_type,
     }
     case IntegralType::interior_facet:
     {
+      // Get indicator for all coefficients that are active in interior
+      // facet integrals
+      for (std::size_t i = 0;
+           i < form.num_integrals(IntegralType::interior_facet); ++i)
+      {
+        for (auto idx : form.active_coeffs(IntegralType::interior_facet, i))
+          active_coefficient[idx] = 1;
+      }
+
       // Iterate over coefficients
       for (std::size_t coeff = 0; coeff < coefficients.size(); ++coeff)
       {
+        if (!active_coefficient[coeff])
+          continue;
+
         auto mesh = coefficients[coeff]->function_space()->mesh();
         std::vector<std::int32_t> facets
             = form.domain(IntegralType::interior_facet, id, *mesh);
+
         std::span<const std::uint32_t> cell_info
             = impl::get_cell_orientation_info(*coefficients[coeff]);
 
@@ -1076,6 +1170,7 @@ void pack_coefficients(const Form<T, U>& form, IntegralType integral_type,
         impl::pack_coefficient_entity(
             c, 2 * cstride, *coefficients[coeff], cell_info, facets, 4,
             [](auto entity) { return entity[0]; }, 2 * offsets[coeff]);
+
         // Pack coefficient ['-']
         impl::pack_coefficient_entity(
             c, 2 * cstride, *coefficients[coeff], cell_info, facets, 4,
@@ -1116,6 +1211,7 @@ Expression<T, U> create_expression(
       tabulate_tensor = nullptr;
   if constexpr (std::is_same_v<T, float>)
     tabulate_tensor = e.tabulate_tensor_float32;
+#ifndef DOLFINX_NO_STDC_COMPLEX_KERNELS
   else if constexpr (std::is_same_v<T, std::complex<float>>)
   {
     tabulate_tensor = reinterpret_cast<void (*)(
@@ -1123,8 +1219,10 @@ Expression<T, U> create_expression(
         const typename scalar_value_type<T>::value_type*, const int*,
         const unsigned char*)>(e.tabulate_tensor_complex64);
   }
+#endif // DOLFINX_NO_STDC_COMPLEX_KERNELS
   else if constexpr (std::is_same_v<T, double>)
     tabulate_tensor = e.tabulate_tensor_float64;
+#ifndef DOLFINX_NO_STDC_COMPLEX_KERNELS
   else if constexpr (std::is_same_v<T, std::complex<double>>)
   {
     tabulate_tensor = reinterpret_cast<void (*)(
@@ -1132,6 +1230,7 @@ Expression<T, U> create_expression(
         const typename scalar_value_type<T>::value_type*, const int*,
         const unsigned char*)>(e.tabulate_tensor_complex128);
   }
+#endif // DOLFINX_NO_STDC_COMPLEX_KERNELS
   else
     throw std::runtime_error("Type not supported.");
 
@@ -1263,8 +1362,7 @@ std::vector<typename U::scalar_type> pack_constants(const U& u)
   for (auto& constant : constants)
   {
     const std::vector<T>& value = constant->value;
-    std::copy(value.begin(), value.end(),
-              std::next(constant_values.begin(), offset));
+    std::ranges::copy(value, std::next(constant_values.begin(), offset));
     offset += value.size();
   }
 
