@@ -45,7 +45,7 @@ XDMFFile::XDMFFile(MPI_Comm comm, const std::filesystem::path& filename,
     _h5_id
         = io::hdf5::open_file(_comm.comm(), hdf5_filename, file_mode, mpi_io);
     assert(_h5_id > 0);
-    LOG(INFO) << "Opened HDF5 file with id \"" << _h5_id << "\"";
+    spdlog::info("Opened HDF5 file with id \"{}\"", _h5_id);
   }
   else
   {
@@ -68,6 +68,12 @@ XDMFFile::XDMFFile(MPI_Comm comm, const std::filesystem::path& filename,
   }
   else if (_file_mode == "w")
   {
+    if (_encoding == Encoding::ASCII and dolfinx::MPI::size(_comm.comm()) > 1)
+    {
+      throw std::runtime_error(
+          "ASCII encoding is not supported for writing files in parallel.");
+    }
+
     _xml_doc->reset();
 
     // Add XDMF node and version attribute
@@ -84,6 +90,12 @@ XDMFFile::XDMFFile(MPI_Comm comm, const std::filesystem::path& filename,
   }
   else if (_file_mode == "a")
   {
+    if (_encoding == Encoding::ASCII and dolfinx::MPI::size(_comm.comm()) > 1)
+    {
+      throw std::runtime_error("ASCII encoding is not supported for appending "
+                               "to files in parallel.");
+    }
+
     if (std::filesystem::exists(_filename))
     {
       // Load XML doc from file
@@ -196,7 +208,7 @@ XDMFFile::read_topology_data(std::string name, std::string xpath) const
   if (!grid_node)
     throw std::runtime_error("<Grid> with name '" + name + "' not found.");
 
-  LOG(INFO) << "Read topology data \"" << name << "\" at \"" << xpath << "\"";
+  spdlog::info("Read topology data \"{}\" at {}", name, xpath);
   return xdmf_mesh::read_topology_data(_comm.comm(), _h5_id, grid_node);
 }
 //-----------------------------------------------------------------------------
@@ -213,7 +225,7 @@ XDMFFile::read_geometry_data(std::string name, std::string xpath) const
   if (!grid_node)
     throw std::runtime_error("<Grid> with name '" + name + "' not found.");
 
-  LOG(INFO) << "Read geometry data \"" << name << "\" at \"" << xpath << "\"";
+  spdlog::info("Read geometry data \"{}\" at {}", name, xpath);
   return xdmf_mesh::read_geometry_data(_comm.comm(), _h5_id, grid_node);
 }
 //-----------------------------------------------------------------------------
@@ -247,8 +259,8 @@ void XDMFFile::write_function(const fem::Function<T, U>& u, double t,
   pugi::xml_node mesh_node = _xml_doc->select_node(mesh_xpath.c_str()).node();
   if (!mesh_node)
   {
-    LOG(WARNING) << "No mesh found at '" << mesh_xpath
-                 << "'. Write mesh before function!";
+    spdlog::warn("No mesh found at '{}'. Write mesh before function!",
+                 mesh_xpath);
   }
 
   const std::string ref_path
@@ -325,7 +337,7 @@ mesh::MeshTags<std::int32_t>
 XDMFFile::read_meshtags(const mesh::Mesh<double>& mesh, std::string name,
                         std::string xpath)
 {
-  LOG(INFO) << "XDMF read meshtags (" << name << ")";
+  spdlog::info("XDMF read meshtags ({})", name);
   pugi::xml_node node = _xml_doc->select_node(xpath.c_str()).node();
   if (!node)
     throw std::runtime_error("XML node '" + xpath + "' not found.");
@@ -360,7 +372,7 @@ XDMFFile::read_meshtags(const mesh::Mesh<double>& mesh, std::string name,
           mesh.geometry().cmap().create_dof_layout(), mesh.geometry().dofmap(),
           mesh::cell_dim(cell_type), entities_span, values);
 
-  LOG(INFO) << "XDMF create meshtags";
+  spdlog::info("XDMF create meshtags");
   std::size_t num_vertices_per_entity = mesh::cell_num_entities(
       mesh::cell_entity_type(mesh.topology()->cell_type(),
                              mesh::cell_dim(cell_type), 0),
