@@ -5,6 +5,7 @@
 // SPDX-License-Identifier:    LGPL-3.0-or-later
 
 #include "xdmf_utils.h"
+#include <algorithm>
 #include <array>
 #include <boost/algorithm/string.hpp>
 #include <boost/lexical_cast.hpp>
@@ -56,12 +57,13 @@ xdmf_utils::get_cell_type(const pugi::xml_node& topology_node)
          {"quadrilateral_9", {"quadrilateral", 2}},
          {"quadrilateral_16", {"quadrilateral", 3}},
          {"hexahedron", {"hexahedron", 1}},
+         {"wedge", {"prism", 1}},
          {"hexahedron_27", {"hexahedron", 2}}};
 
   // Convert XDMF cell type string to DOLFINx cell type string
   std::string cell_type = type_attr.as_string();
-  std::transform(cell_type.begin(), cell_type.end(), cell_type.begin(),
-                 [](unsigned char c) { return std::tolower(c); });
+  std::ranges::transform(cell_type, cell_type.begin(),
+                         [](auto c) { return std::tolower(c); });
   auto it = xdmf_to_dolfin.find(cell_type);
   if (it == xdmf_to_dolfin.end())
   {
@@ -227,7 +229,7 @@ xdmf_utils::distribute_entity_data(
     std::span<const T> data)
 {
   assert(entities.extent(0) == data.size());
-  LOG(INFO) << "XDMF distribute entity data";
+  spdlog::info("XDMF distribute entity data");
   mesh::CellType cell_type = topology.cell_type();
 
   // Get layout of dofs on 0th cell entity of dimension entity_dim
@@ -270,7 +272,7 @@ xdmf_utils::distribute_entity_data(
       std::span entity(entities_v.data() + e * num_vert_per_e, num_vert_per_e);
       for (std::size_t i = 0; i < num_vert_per_e; ++i)
         entity[i] = entities(e, entity_vertex_dofs[i]);
-      std::sort(entity.begin(), entity.end());
+      std::ranges::sort(entity);
     }
 
     std::array shape{entities.extent(0), num_vert_per_e};
@@ -301,8 +303,8 @@ xdmf_utils::distribute_entity_data(
     }
     std::vector<int> perm(dest0.size());
     std::iota(perm.begin(), perm.end(), 0);
-    std::sort(perm.begin(), perm.end(),
-              [&dest0](auto x0, auto x1) { return dest0[x0] < dest0[x1]; });
+    std::ranges::sort(perm, [&dest0](auto x0, auto x1)
+                      { return dest0[x0] < dest0[x1]; });
 
     // Note: dest[perm[i]] is ordered with increasing i
     // Build list of neighbour dest ranks and count number of entities to
@@ -330,7 +332,7 @@ xdmf_utils::distribute_entity_data(
     // Determine src ranks. Sort ranks so that ownership determination is
     // deterministic for a given number of ranks.
     std::vector<int> src = dolfinx::MPI::compute_graph_edges_nbx(comm, dest);
-    std::sort(src.begin(), src.end());
+    std::ranges::sort(src);
 
     // Create neighbourhood communicator for sending data to post
     // offices
@@ -399,12 +401,12 @@ xdmf_utils::distribute_entity_data(
   {
     int size = dolfinx::MPI::size(comm);
     std::vector<std::pair<int, std::int64_t>> dest_to_index;
-    std::transform(
-        indices.begin(), indices.end(), std::back_inserter(dest_to_index),
+    std::ranges::transform(
+        indices, std::back_inserter(dest_to_index),
         [size, num_nodes](auto n) {
           return std::pair(dolfinx::MPI::index_owner(size, n, num_nodes), n);
         });
-    std::sort(dest_to_index.begin(), dest_to_index.end());
+    std::ranges::sort(dest_to_index);
 
     // Build list of neighbour dest ranks and count number of indices to
     // send to each post office
@@ -431,7 +433,7 @@ xdmf_utils::distribute_entity_data(
     // Determine src ranks. Sort ranks so that ownership determination is
     // deterministic for a given number of ranks.
     std::vector<int> src = dolfinx::MPI::compute_graph_edges_nbx(comm, dest);
-    std::sort(src.begin(), src.end());
+    std::ranges::sort(src);
 
     // Create neighbourhood communicator for sending data to post offices
     MPI_Comm comm0;
@@ -457,9 +459,8 @@ xdmf_utils::distribute_entity_data(
     // Prepare send buffer
     std::vector<std::int64_t> send_buffer;
     send_buffer.reserve(indices.size());
-    std::transform(dest_to_index.begin(), dest_to_index.end(),
-                   std::back_inserter(send_buffer),
-                   [](auto x) { return x.second; });
+    std::ranges::transform(dest_to_index, std::back_inserter(send_buffer),
+                           [](auto x) { return x.second; });
 
     std::vector<std::int64_t> recv_buffer(recv_disp.back());
     err = MPI_Neighbor_alltoallv(send_buffer.data(), num_items_send.data(),
@@ -586,7 +587,7 @@ xdmf_utils::distribute_entity_data(
            std::span<const int> cell_vertex_dofs, auto entities_data,
            std::span<const T> entities_values)
   {
-    LOG(INFO) << "XDMF build map";
+    spdlog::info("XDMF build map");
     auto c_to_v = topology.connectivity(topology.dim(), 0);
     if (!c_to_v)
       throw std::runtime_error("Missing cell-vertex connectivity.");
@@ -699,4 +700,4 @@ xdmf_utils::distribute_entity_data(
         MDSPAN_IMPL_STANDARD_NAMESPACE::dextents<std::size_t, 2>>,
     std::span<const std::complex<float>>);
 
-/// @endcond
\ No newline at end of file
+/// @endcond
