@@ -11,6 +11,7 @@
 #include "SparsityPattern.h"
 #include "Vector.h"
 #include "utils.h"
+#include <algorithm>
 #include <dolfinx/common/IndexMap.h>
 #include <dolfinx/common/MPI.h>
 #include <dolfinx/common/Timer.h>
@@ -38,9 +39,9 @@ void la::petsc::error(int error_code, std::string filename,
   PetscErrorMessage(error_code, &desc, nullptr);
 
   // Log detailed error info
-  DLOG(INFO) << "PETSc error in '" << filename.c_str() << "', '"
-             << petsc_function.c_str() << "'";
-  DLOG(INFO) << "PETSc error code '" << error_code << "' (" << desc << ".";
+  spdlog::info("PETSc error in '{}', '{}'", filename.c_str(),
+               petsc_function.c_str());
+  spdlog::info("PETSc error code '{}' '{}'", error_code, desc);
   throw std::runtime_error("Failed to successfully call PETSc function '"
                            + petsc_function + "'. PETSc error code is: "
                            + std ::to_string(error_code) + ", "
@@ -57,7 +58,7 @@ la::petsc::create_vectors(MPI_Comm comm,
     VecCreateMPI(comm, x[i].size(), PETSC_DETERMINE, &v[i]);
     PetscScalar* data;
     VecGetArray(v[i], &data);
-    std::copy(x[i].begin(), x[i].end(), data);
+    std::ranges::copy(x[i], data);
     VecRestoreArray(v[i], &data);
   }
 
@@ -549,8 +550,9 @@ Vec petsc::Operator::create_vector(std::size_t dim) const
   }
   else
   {
-    LOG(ERROR) << "Cannot initialize PETSc vector to match PETSc matrix. "
-               << "Dimension must be 0 or 1, not " << dim;
+    spdlog::error("Cannot initialize PETSc vector to match PETSc matrix. "
+                  "Dimension must be 0 or 1, not {}",
+                  dim);
     throw std::runtime_error("Invalid dimension");
   }
 
@@ -697,7 +699,7 @@ int petsc::KrylovSolver::solve(Vec x, const Vec b, bool transpose) const
   PetscErrorCode ierr;
 
   // Solve linear system
-  LOG(INFO) << "PETSc Krylov solver starting to solve system.";
+  spdlog::info("PETSc Krylov solver starting to solve system.");
 
   // Solve system
   if (!transpose)
