@@ -49,8 +49,7 @@ determine_sharing_ranks(MPI_Comm comm, std::span<const std::int64_t> indices)
   std::int64_t global_range = 0;
   {
     std::int64_t max_index
-        = indices.empty() ? 0
-                          : *std::max_element(indices.begin(), indices.end());
+        = indices.empty() ? 0 : *std::ranges::max_element(indices);
     MPI_Allreduce(&max_index, &global_range, 1, MPI_INT64_T, MPI_MAX, comm);
     global_range += 1;
   }
@@ -64,7 +63,7 @@ determine_sharing_ranks(MPI_Comm comm, std::span<const std::int64_t> indices)
       int dest = dolfinx::MPI::index_owner(size, idx, global_range);
       dest_to_index.push_back({dest, static_cast<int>(dest_to_index.size())});
     }
-    std::sort(dest_to_index.begin(), dest_to_index.end());
+    std::ranges::sort(dest_to_index);
   }
 
   // Build list of neighbour dest ranks and count number of indices to
@@ -92,7 +91,7 @@ determine_sharing_ranks(MPI_Comm comm, std::span<const std::int64_t> indices)
   // Determine src ranks. Sort ranks so that ownership determination is
   // deterministic for a given number of ranks.
   std::vector<int> src = dolfinx::MPI::compute_graph_edges_nbx(comm, dest);
-  std::sort(src.begin(), src.end());
+  std::ranges::sort(src);
 
   // Create neighbourhood communicator for sending data to post offices
   MPI_Comm neigh_comm0;
@@ -141,7 +140,7 @@ determine_sharing_ranks(MPI_Comm comm, std::span<const std::int64_t> indices)
     for (std::int32_t i = recv_disp0[p]; i < recv_disp0[p + 1]; ++i)
       indices_list.push_back({recv_buffer0[i], i, int(p)});
   }
-  std::sort(indices_list.begin(), indices_list.end());
+  std::ranges::sort(indices_list);
 
   // Find which ranks have each index
   std::vector<std::int32_t> num_items_per_dest1(recv_disp0.size() - 1, 0);
@@ -304,11 +303,12 @@ std::array<std::vector<std::int64_t>, 2> vertex_ownership_groups(
                       [](std::size_t s, auto& v) { return s + v.size(); }));
   for (auto c : cells_owned)
     local_vertex_set.insert(local_vertex_set.end(), c.begin(), c.end());
-  dolfinx::radix_sort(std::span(local_vertex_set));
-  local_vertex_set.erase(
-      std::unique(local_vertex_set.begin(), local_vertex_set.end()),
-      local_vertex_set.end());
 
+  {
+    dolfinx::radix_sort(local_vertex_set);
+    auto [unique_end, range_end] = std::ranges::unique(local_vertex_set);
+    local_vertex_set.erase(unique_end, range_end);
+  }
   // Build set of ghost cell vertices (attached to a ghost cell)
   std::vector<std::int64_t> ghost_vertex_set;
   ghost_vertex_set.reserve(
@@ -316,32 +316,30 @@ std::array<std::vector<std::int64_t>, 2> vertex_ownership_groups(
                       [](std::size_t s, auto& v) { return s + v.size(); }));
   for (auto c : cells_ghost)
     ghost_vertex_set.insert(ghost_vertex_set.end(), c.begin(), c.end());
-  dolfinx::radix_sort(std::span(ghost_vertex_set));
-  ghost_vertex_set.erase(
-      std::unique(ghost_vertex_set.begin(), ghost_vertex_set.end()),
-      ghost_vertex_set.end());
 
+  {
+    dolfinx::radix_sort(ghost_vertex_set);
+    auto [unique_end, range_end] = std::ranges::unique(ghost_vertex_set);
+    ghost_vertex_set.erase(unique_end, range_end);
+  }
   // Build difference 1: Vertices attached only to owned cells, and
   // therefore owned by this rank
   std::vector<std::int64_t> owned_vertices;
-  std::set_difference(local_vertex_set.begin(), local_vertex_set.end(),
-                      boundary_vertices.begin(), boundary_vertices.end(),
-                      std::back_inserter(owned_vertices));
+  std::ranges::set_difference(local_vertex_set, boundary_vertices,
+                              std::back_inserter(owned_vertices));
 
   // Build difference 2: Vertices attached only to ghost cells, and
   // therefore not owned by this rank
   std::vector<std::int64_t> unowned_vertices;
-  std::set_difference(ghost_vertex_set.begin(), ghost_vertex_set.end(),
-                      local_vertex_set.begin(), local_vertex_set.end(),
-                      std::back_inserter(unowned_vertices));
+  std::ranges::set_difference(ghost_vertex_set, local_vertex_set,
+                              std::back_inserter(unowned_vertices));
 
   // TODO Check this in debug mode only?
   // Sanity check
   // No vertices in unowned should also be in boundary...
   std::vector<std::int64_t> unowned_vertices_in_error;
-  std::set_intersection(unowned_vertices.begin(), unowned_vertices.end(),
-                        boundary_vertices.begin(), boundary_vertices.end(),
-                        std::back_inserter(unowned_vertices_in_error));
+  std::ranges::set_intersection(unowned_vertices, boundary_vertices,
+                                std::back_inserter(unowned_vertices_in_error));
 
   if (!unowned_vertices_in_error.empty())
   {
@@ -389,10 +387,18 @@ exchange_indexing(MPI_Comm comm, std::span<const std::int64_t> indices,
     else
       src.push_back(ranks.front());
   }
-  std::sort(src.begin(), src.end());
-  src.erase(std::unique(src.begin(), src.end()), src.end());
-  std::sort(dest.begin(), dest.end());
-  dest.erase(std::unique(dest.begin(), dest.end()), dest.end());
+
+  {
+    std::ranges::sort(src);
+    auto [unique_end, range_end] = std::ranges::unique(src);
+    src.erase(unique_end, range_end);
+  }
+
+  {
+    std::ranges::sort(dest);
+    auto [unique_end, range_end] = std::ranges::unique(dest);
+    dest.erase(unique_end, range_end);
+  }
 
   // Pack send data. Use std::vector<std::vector>> since size will be
   // modest (equal to number of neighbour ranks)
@@ -406,8 +412,7 @@ exchange_indexing(MPI_Comm comm, std::span<const std::int64_t> indices,
     {
       // Get local vertex index
       std::int64_t idx_old = indices[i];
-      auto local_it = std::lower_bound(global_indices.begin(),
-                                       global_indices.end(), idx_old);
+      auto local_it = std::ranges::lower_bound(global_indices, idx_old);
       assert(local_it != global_indices.end() and *local_it == idx_old);
       std::size_t pos = std::distance(global_indices.begin(), local_it);
       std::int64_t idx_new = local_indices[pos] + offset;
@@ -417,7 +422,7 @@ exchange_indexing(MPI_Comm comm, std::span<const std::int64_t> indices,
       for (std::size_t j = 1; j < ranks.size(); ++j)
       {
         // Find rank on the neighborhood comm
-        auto it = std::lower_bound(dest.begin(), dest.end(), ranks[j]);
+        auto it = std::ranges::lower_bound(dest, ranks[j]);
         assert(it != dest.end() and *it == ranks[j]);
         int neighbor = std::distance(dest.begin(), it);
 
@@ -440,9 +445,8 @@ exchange_indexing(MPI_Comm comm, std::span<const std::int64_t> indices,
     // Prepare send sizes and send displacements
     std::vector<int> send_sizes;
     send_sizes.reserve(dest.size());
-    std::transform(send_buffer.begin(), send_buffer.end(),
-                   std::back_inserter(send_sizes),
-                   [](auto& x) { return x.size(); });
+    std::ranges::transform(send_buffer, std::back_inserter(send_sizes),
+                           [](auto& x) { return x.size(); });
     std::vector<int> send_disp(dest.size() + 1, 0);
     std::partial_sum(send_sizes.begin(), send_sizes.end(),
                      std::next(send_disp.begin()));
@@ -539,19 +543,18 @@ std::vector<std::array<std::int64_t, 3>> exchange_ghost_indexing(
     // Build list of (owner rank, index) pairs for each ghost index, and
     // sort
     std::vector<std::pair<int, std::int64_t>> owner_to_ghost;
-    std::transform(map0.ghosts().begin(), map0.ghosts().end(),
-                   map0.owners().begin(), std::back_inserter(owner_to_ghost),
-                   [](auto idx, auto r) -> std::pair<int, std::int64_t>
-                   { return {r, idx}; });
-    std::sort(owner_to_ghost.begin(), owner_to_ghost.end());
+    std::ranges::transform(map0.ghosts(), map0.owners(),
+                           std::back_inserter(owner_to_ghost),
+                           [](auto idx, auto r) -> std::pair<int, std::int64_t>
+                           { return {r, idx}; });
+    std::ranges::sort(owner_to_ghost);
 
     // Build send buffer (the second component of each pair in
     // owner_to_ghost) to send to rank that owns the index
     std::vector<std::int64_t> send_buffer;
     send_buffer.reserve(owner_to_ghost.size());
-    std::transform(owner_to_ghost.begin(), owner_to_ghost.end(),
-                   std::back_inserter(send_buffer),
-                   [](auto x) { return x.second; });
+    std::ranges::transform(owner_to_ghost, std::back_inserter(send_buffer),
+                           [](auto x) { return x.second; });
 
     // Compute send sizes and displacements
     std::vector<int> send_sizes, send_disp{0};
@@ -602,17 +605,16 @@ std::vector<std::array<std::int64_t, 3>> exchange_ghost_indexing(
                                vertices.end());
       }
 
-      std::sort(shared_vertices.begin(), shared_vertices.end());
-      shared_vertices.erase(
-          std::unique(shared_vertices.begin(), shared_vertices.end()),
-          shared_vertices.end());
+      std::ranges::sort(shared_vertices);
+      auto [unique_end, range_end] = std::ranges::unique(shared_vertices);
+      shared_vertices.erase(unique_end, range_end);
     }
   }
 
   // Compute send sizes and offsets
   std::vector<int> send_sizes(dest.size());
-  std::transform(shared_vertices_fwd.begin(), shared_vertices_fwd.end(),
-                 send_sizes.begin(), [](auto& x) { return 3 * x.size(); });
+  std::ranges::transform(shared_vertices_fwd, send_sizes.begin(),
+                         [](auto& x) { return 3 * x.size(); });
   std::vector<int> send_disp(dest.size() + 1);
   std::partial_sum(send_sizes.begin(), send_sizes.end(),
                    std::next(send_disp.begin()));
@@ -637,8 +639,8 @@ std::vector<std::array<std::int64_t, 3>> exchange_ghost_indexing(
       for (auto vertex_old : vertices_old)
       {
         // Find new vertex index and determine owning rank
-        auto it = std::lower_bound(
-            global_local_entities1.begin(), global_local_entities1.end(),
+        auto it = std::ranges::lower_bound(
+            global_local_entities1,
             std::pair<std::int64_t, std::int32_t>(vertex_old, 0),
             [](auto& a, auto& b) { return a.first < b.first; });
         assert(it != global_local_entities1.end());
@@ -672,8 +674,9 @@ std::vector<std::array<std::int64_t, 3>> exchange_ghost_indexing(
   data.reserve(recv_buffer.size() / 3);
   for (std::size_t i = 0; i < recv_buffer.size(); i += 3)
     data.push_back({recv_buffer[i], recv_buffer[i + 1], recv_buffer[i + 2]});
-  std::sort(data.begin(), data.end());
-  data.erase(std::unique(data.begin(), data.end()), data.end());
+  std::ranges::sort(data);
+  auto [unique_end, range_end] = std::ranges::unique(data);
+  data.erase(unique_end, range_end);
 
   MPI_Comm_free(&comm);
 
@@ -697,10 +700,9 @@ std::vector<std::int32_t> convert_to_local_indexing(
   std::transform(g.begin(), std::next(g.begin(), data.size()), data.begin(),
                  [&global_to_local](auto i)
                  {
-                   auto it = std::lower_bound(
-                       global_to_local.begin(), global_to_local.end(),
-                       std::pair<std::int64_t, std::int32_t>(i, 0),
-                       [](auto& a, auto& b) { return a.first < b.first; });
+                   auto it = std::ranges::lower_bound(
+                       global_to_local, i, std::ranges::less(),
+                       [](auto& e) { return e.first; });
                    assert(it != global_to_local.end());
                    assert(it->first == i);
                    return it->second;
@@ -747,7 +749,7 @@ Topology::Topology(MPI_Comm comm, const std::vector<CellType>& cell_types)
   assert(!cell_types.empty());
   std::int8_t tdim = cell_dim(cell_types[0]);
   assert(tdim > 0);
-  for (auto ct : cell_types)
+  for ([[maybe_unused]] auto ct : cell_types)
     assert(cell_dim(ct) == tdim);
   _interprocess_facets.resize(1);
 
@@ -857,7 +859,7 @@ std::int32_t Topology::create_entities(int dim)
     // Store boundary facets
     if (dim == this->dim() - 1)
     {
-      std::sort(interprocess_entities.begin(), interprocess_entities.end());
+      std::ranges::sort(interprocess_entities);
       assert(index < _interprocess_facets.size());
       _interprocess_facets[index] = std::move(interprocess_entities);
     }
@@ -1041,7 +1043,7 @@ Topology mesh::create_topology(
   assert(ghost_owners.size() == cells.size());
   assert(original_cell_index.size() == cells.size());
 
-  LOG(INFO) << "Create topology (generalised)";
+  spdlog::info("Create topology (generalised)");
   // Check cell data consistency and compile spans of owned and ghost cells
   std::vector<std::int32_t> num_local_cells(cell_type.size());
   std::vector<std::span<const std::int64_t>> owned_cells;
@@ -1092,12 +1094,12 @@ Topology mesh::create_topology(
       else
         unowned_vertices.push_back(global_index);
     }
-    dolfinx::radix_sort(std::span(unowned_vertices));
+    dolfinx::radix_sort(unowned_vertices);
 
     // Add owned but shared vertices to owned_vertices, and sort
     owned_vertices.insert(owned_vertices.end(), owned_shared_vertices.begin(),
                           owned_shared_vertices.end());
-    dolfinx::radix_sort(std::span(owned_vertices));
+    dolfinx::radix_sort(owned_vertices);
   }
 
   // Number all owned vertices, iterating over vertices cell-wise
@@ -1108,8 +1110,7 @@ Topology mesh::create_topology(
     {
       for (auto vtx : cells[i])
       {
-        auto it = std::lower_bound(owned_vertices.begin(), owned_vertices.end(),
-                                   vtx);
+        auto it = std::ranges::lower_bound(owned_vertices, vtx);
         if (it != owned_vertices.end() and *it == vtx)
         {
           std::size_t pos = std::distance(owned_vertices.begin(), it);
@@ -1161,8 +1162,7 @@ Topology mesh::create_topology(
     for (std::size_t i = 0; i < unowned_vertex_data.size(); i += 3)
     {
       const std::int64_t idx_global = unowned_vertex_data[i];
-      auto it = std::lower_bound(unowned_vertices.begin(),
-                                 unowned_vertices.end(), idx_global);
+      auto it = std::ranges::lower_bound(unowned_vertices, idx_global);
       assert(it != unowned_vertices.end() and *it == idx_global);
       std::size_t pos = std::distance(unowned_vertices.begin(), it);
       assert(local_vertex_indices_unowned[pos] < 0);
@@ -1177,20 +1177,15 @@ Topology mesh::create_topology(
           global_to_local_vertices;
       global_to_local_vertices.reserve(owned_vertices.size()
                                        + unowned_vertices.size());
-      std::transform(owned_vertices.begin(), owned_vertices.end(),
-                     local_vertex_indices.begin(),
-                     std::back_inserter(global_to_local_vertices),
-                     [](auto idx0, auto idx1) {
-                       return std::pair<std::int64_t, std::int32_t>(idx0, idx1);
-                     });
-      std::transform(unowned_vertices.begin(), unowned_vertices.end(),
-                     local_vertex_indices_unowned.begin(),
-                     std::back_inserter(global_to_local_vertices),
-                     [](auto idx0, auto idx1) {
-                       return std::pair<std::int64_t, std::int32_t>(idx0, idx1);
-                     });
-      std::sort(global_to_local_vertices.begin(),
-                global_to_local_vertices.end());
+      std::ranges::transform(
+          owned_vertices, local_vertex_indices,
+          std::back_inserter(global_to_local_vertices), [](auto idx0, auto idx1)
+          { return std::pair<std::int64_t, std::int32_t>(idx0, idx1); });
+      std::ranges::transform(
+          unowned_vertices, local_vertex_indices_unowned,
+          std::back_inserter(global_to_local_vertices), [](auto idx0, auto idx1)
+          { return std::pair<std::int64_t, std::int32_t>(idx0, idx1); });
+      std::ranges::sort(global_to_local_vertices);
 
       // Send (from the ghost cell owner) and receive global indices for
       // ghost vertices that are not on the process boundary. Data is
@@ -1215,8 +1210,7 @@ Topology mesh::create_topology(
       for (auto& data : recv_data)
       {
         std::int64_t global_idx_old = data[0];
-        auto it0 = std::lower_bound(unowned_vertices.begin(),
-                                    unowned_vertices.end(), global_idx_old);
+        auto it0 = std::ranges::lower_bound(unowned_vertices, global_idx_old);
         if (it0 != unowned_vertices.end() and *it0 == global_idx_old)
         {
           if (std::size_t pos = std::distance(unowned_vertices.begin(), it0);
@@ -1237,19 +1231,17 @@ Topology mesh::create_topology(
   std::vector<std::pair<std::int64_t, std::int32_t>> global_to_local_vertices;
   global_to_local_vertices.reserve(owned_vertices.size()
                                    + unowned_vertices.size());
-  std::transform(
-      owned_vertices.begin(), owned_vertices.end(),
-      local_vertex_indices.begin(),
+  std::ranges::transform(
+      owned_vertices, local_vertex_indices,
       std::back_inserter(global_to_local_vertices),
       [](auto idx0, auto idx1) -> std::pair<std::int64_t, std::int32_t>
       { return {idx0, idx1}; });
-  std::transform(
-      unowned_vertices.begin(), unowned_vertices.end(),
-      local_vertex_indices_unowned.begin(),
+  std::ranges::transform(
+      unowned_vertices, local_vertex_indices_unowned,
       std::back_inserter(global_to_local_vertices),
       [](auto idx0, auto idx1) -> std::pair<std::int64_t, std::int32_t>
       { return {idx0, idx1}; });
-  std::sort(global_to_local_vertices.begin(), global_to_local_vertices.end());
+  std::ranges::sort(global_to_local_vertices);
 
   std::vector<std::vector<std::int32_t>> _cells_local_idx(cells.size());
   for (std::size_t i = 0; i < cell_type.size(); ++i)
@@ -1281,8 +1273,10 @@ Topology mesh::create_topology(
     // Build list of ranks that own vertices that are ghosted by this
     // rank (out edges)
     std::vector<int> src = ghost_vertex_owners;
-    dolfinx::radix_sort(std::span(src));
-    src.erase(std::unique(src.begin(), src.end()), src.end());
+    dolfinx::radix_sort(src);
+    auto [unique_end, range_end] = std::ranges::unique(src);
+    src.erase(unique_end, range_end);
+
     dest = dolfinx::MPI::compute_graph_edges_nbx(comm, src);
   }
 
@@ -1327,7 +1321,7 @@ mesh::create_topology(MPI_Comm comm, std::span<const std::int64_t> cells,
                       std::span<const int> ghost_owners, CellType cell_type,
                       std::span<const std::int64_t> boundary_vertices)
 {
-  LOG(INFO) << "Create topology (single cell type)";
+  spdlog::info("Create topology (single cell type)");
 
   return create_topology(comm, {cell_type}, {cells}, {original_cell_index},
                          {ghost_owners}, boundary_vertices);
@@ -1348,9 +1342,10 @@ mesh::create_subtopology(const Topology& topology, int dim,
   {
     // FIXME Make this an input requirement?
     std::vector<std::int32_t> _entities(entities.begin(), entities.end());
-    std::sort(_entities.begin(), _entities.end());
-    _entities.erase(std::unique(_entities.begin(), _entities.end()),
-                    _entities.end());
+    std::ranges::sort(_entities);
+    auto [unique_end, range_end] = std::ranges::unique(_entities);
+    _entities.erase(unique_end, range_end);
+
     auto [_submap, _subentities]
         = common::create_sub_index_map(*topology.index_map(dim), _entities);
     submap = std::make_shared<common::IndexMap>(std::move(_submap));
@@ -1430,7 +1425,7 @@ std::vector<std::int32_t>
 mesh::entities_to_index(const Topology& topology, int dim,
                         std::span<const std::int32_t> entities)
 {
-  LOG(INFO) << "Build list of mesh entity indices from the entity vertices.";
+  spdlog::info("Build list of mesh entity indices from the entity vertices.");
 
   // Tagged entity topological dimension
   auto map_e = topology.index_map(dim);
@@ -1454,8 +1449,8 @@ mesh::entities_to_index(const Topology& topology, int dim,
   for (int e = 0; e < num_entities_mesh; ++e)
   {
     auto vertices = e_to_v->links(e);
-    std::copy(vertices.begin(), vertices.end(), key.begin());
-    std::sort(key.begin(), key.end());
+    std::ranges::copy(vertices, key.begin());
+    std::ranges::sort(key);
     auto ins = entity_key_to_index.insert({key, e});
     if (!ins.second)
       throw std::runtime_error("Duplicate mesh entity detected.");
@@ -1470,8 +1465,8 @@ mesh::entities_to_index(const Topology& topology, int dim,
   for (std::size_t e = 0; e < entities.size(); e += num_vertices_per_entity)
   {
     auto v = entities.subspan(e, num_vertices_per_entity);
-    std::copy(v.begin(), v.end(), vertices.begin());
-    std::sort(vertices.begin(), vertices.end());
+    std::ranges::copy(v, vertices.begin());
+    std::ranges::sort(vertices);
     if (auto it = entity_key_to_index.find(vertices);
         it != entity_key_to_index.end())
     {
