@@ -39,13 +39,14 @@ namespace
 template <typename U>
 graph::AdjacencyList<int> create_adj_list(U& data, std::int32_t size)
 {
-  std::sort(data.begin(), data.end());
-  data.erase(std::unique(data.begin(), data.end()), data.end());
+  std::ranges::sort(data);
+  auto [unique_end, range_end] = std::ranges::unique(data);
+  data.erase(unique_end, range_end);
 
   std::vector<int> array;
   array.reserve(data.size());
-  std::transform(data.begin(), data.end(), std::back_inserter(array),
-                 [](auto x) { return x.second; });
+  std::ranges::transform(data, std::back_inserter(array),
+                         [](auto x) { return x.second; });
 
   std::vector<std::int32_t> offsets{0};
   offsets.reserve(size + 1);
@@ -94,7 +95,7 @@ int get_ownership(const U& processes, const V& vertices)
 /// @param[in] num_entities_per_cell Number of entities per cell
 /// @param[in] entity_index Initial numbering for each row in
 /// entity_list
-/// @returns Local indices and index map
+/// @returns Local indices, the index map and shared entities
 std::tuple<std::vector<int>, common::IndexMap, std::vector<std::int32_t>>
 get_local_indexing(MPI_Comm comm, const common::IndexMap& vertex_map,
                    std::span<const std::int32_t> entity_list,
@@ -114,7 +115,7 @@ get_local_indexing(MPI_Comm comm, const common::IndexMap& vertex_map,
 
   // Find the maximum entity index, hence the number of entities
   std::int32_t entity_count = 0;
-  if (auto mx = std::max_element(entity_index.begin(), entity_index.end());
+  if (auto mx = std::ranges::max_element(entity_index);
       mx != entity_index.end())
   {
     entity_count = *mx + 1;
@@ -129,8 +130,9 @@ get_local_indexing(MPI_Comm comm, const common::IndexMap& vertex_map,
   // Create unique list of ranks that share vertices (owners of)
   std::vector<int> ranks(vertex_ranks.array().begin(),
                          vertex_ranks.array().end());
-  std::sort(ranks.begin(), ranks.end());
-  ranks.erase(std::unique(ranks.begin(), ranks.end()), ranks.end());
+  std::ranges::sort(ranks);
+  auto [unique_end, range_end] = std::ranges::unique(ranks);
+  ranks.erase(unique_end, range_end);
 
   MPI_Comm neighbor_comm;
   MPI_Dist_graph_create_adjacent(
@@ -168,7 +170,7 @@ get_local_indexing(MPI_Comm comm, const common::IndexMap& vertex_map,
         entity_ranks.insert(entity_ranks.end(), vertex_ranks.links(v).begin(),
                             vertex_ranks.links(v).end());
       }
-      std::sort(entity_ranks.begin(), entity_ranks.end());
+      std::ranges::sort(entity_ranks);
 
       // If the number of vertices shared with a rank is
       // 'num_vertices_per_e', then add entity data to the send buffer
@@ -180,7 +182,7 @@ get_local_indexing(MPI_Comm comm, const common::IndexMap& vertex_map,
         if (std::distance(it, it1) == num_vertices_per_e)
         {
           vertex_map.local_to_global(entity, vglobal);
-          std::sort(vglobal.begin(), vglobal.end());
+          std::ranges::sort(vglobal);
           entity_to_local_idx.insert(entity_to_local_idx.end(), vglobal.begin(),
                                      vglobal.end());
           entity_to_local_idx.push_back(*entity_idx);
@@ -188,7 +190,7 @@ get_local_indexing(MPI_Comm comm, const common::IndexMap& vertex_map,
           // Only send entities that are not known to be ghosts
           if (ghost_status[*entity_idx] != 1)
           {
-            auto itr_local = std::lower_bound(ranks.begin(), ranks.end(), *it);
+            auto itr_local = std::ranges::lower_bound(ranks, *it);
             assert(itr_local != ranks.end() and *itr_local == *it);
             const int r = std::distance(ranks.begin(), itr_local);
 
@@ -205,24 +207,20 @@ get_local_indexing(MPI_Comm comm, const common::IndexMap& vertex_map,
 
     perm.resize(entity_to_local_idx.size() / (num_vertices_per_e + 1));
     std::iota(perm.begin(), perm.end(), 0);
-    std::sort(perm.begin(), perm.end(),
-              [&entities = entity_to_local_idx,
-               shape = num_vertices_per_e + 1](auto e0, auto e1)
-              {
-                auto it0 = std::next(entities.begin(), e0 * shape);
-                auto it1 = std::next(entities.begin(), e1 * shape);
-                return std::lexicographical_compare(it0, std::next(it0, shape),
-                                                    it1, std::next(it1, shape));
-              });
-    perm.erase(std::unique(perm.begin(), perm.end(),
-                           [&entities = entity_to_local_idx,
-                            shape = num_vertices_per_e + 1](auto e0, auto e1)
-                           {
-                             auto it0 = std::next(entities.begin(), e0 * shape);
-                             auto it1 = std::next(entities.begin(), e1 * shape);
-                             return std::equal(it0, std::next(it0, shape), it1);
-                           }),
-               perm.end());
+
+    auto range_by_index = [&, shape = num_vertices_per_e + 1](auto e)
+    {
+      auto begin = std::next(entity_to_local_idx.begin(), e * shape);
+      return std::ranges::subrange(begin, std::next(begin, shape));
+    };
+
+    std::ranges::sort(perm, std::ranges::lexicographical_compare,
+                      range_by_index);
+
+    auto [unique_end, range_end]
+        = std::ranges::unique(perm, std::ranges::equal, range_by_index);
+
+    perm.erase(unique_end, range_end);
   }
 
   // Get shared entities of this dimension, and also match up an index
@@ -306,9 +304,8 @@ get_local_indexing(MPI_Comm comm, const common::IndexMap& vertex_map,
           shared_entities_data.push_back({idx, ranks[r]});
           shared_entities_data.push_back({idx, mpi_rank});
           recv_index.push_back(idx);
-          std::transform(
-              entity.begin(), entity.end(),
-              std::back_inserter(shared_entity_to_global_vertices_data),
+          std::ranges::transform(
+              entity, std::back_inserter(shared_entity_to_global_vertices_data),
               [idx](auto v) -> std::pair<std::int32_t, std::int64_t>
               { return {idx, v}; });
         }
@@ -363,15 +360,13 @@ get_local_indexing(MPI_Comm comm, const common::IndexMap& vertex_map,
     }
     num_local = c;
 
-    std::transform(local_index.cbegin(), local_index.cend(),
-                   local_index.begin(),
-                   [&c](auto index) { return index == -1 ? c++ : index; });
+    std::ranges::transform(local_index, local_index.begin(), [&c](auto index)
+                           { return index == -1 ? c++ : index; });
     assert(c == entity_count);
 
     // Convert interprocess entities to local_index
-    std::transform(interprocess_entities.cbegin(), interprocess_entities.cend(),
-                   interprocess_entities.begin(),
-                   [&local_index](std::int32_t i) { return local_index[i]; });
+    std::ranges::transform(interprocess_entities, interprocess_entities.begin(),
+                           [&local_index](auto i) { return local_index[i]; });
   }
 
   //---------
@@ -389,25 +384,20 @@ get_local_indexing(MPI_Comm comm, const common::IndexMap& vertex_map,
     std::vector<std::int64_t> send_global_index_data;
     for (const auto& indices : send_index)
     {
-      std::transform(indices.cbegin(), indices.cend(),
-                     std::back_inserter(send_global_index_data),
-                     [&local_index, size = num_local,
-                      offset = local_offset](auto idx) -> std::int64_t
-                     {
-                       // If not in our local range, send -1.
-                       return local_index[idx] < size
-                                  ? offset + local_index[idx]
-                                  : -1;
-                     });
+      std::ranges::transform(
+          indices, std::back_inserter(send_global_index_data),
+          [&local_index, size = num_local,
+           offset = local_offset](auto idx) -> std::int64_t
+          {
+            // If not in our local range, send -1.
+            return local_index[idx] < size ? offset + local_index[idx] : -1;
+          });
     }
 
     // Transform send/receive sizes and displacements for scalar send
     for (auto x : {&send_sizes, &send_disp, &recv_sizes, &recv_disp})
-    {
-      std::transform(x->begin(), x->end(), x->begin(),
-                     [num_vertices_per_e](auto a)
-                     { return a / num_vertices_per_e; });
-    }
+      std::ranges::transform(*x, x->begin(), [num_vertices_per_e](auto a)
+                             { return a / num_vertices_per_e; });
 
     recv_data.resize(recv_disp.back());
     MPI_Neighbor_alltoallv(send_global_index_data.data(), send_sizes.data(),
@@ -440,9 +430,9 @@ get_local_indexing(MPI_Comm comm, const common::IndexMap& vertex_map,
 
   // Create map from initial numbering to new local indices
   std::vector<std::int32_t> new_entity_index(entity_index.size());
-  std::transform(entity_index.begin(), entity_index.end(),
-                 new_entity_index.begin(),
-                 [&local_index](auto index) { return local_index[index]; });
+  std::ranges::transform(entity_index, new_entity_index.begin(),
+                         [&local_index](auto index)
+                         { return local_index[index]; });
 
   return {std::move(new_entity_index), std::move(index_map),
           std::move(interprocess_entities)};
@@ -531,10 +521,42 @@ compute_entities_by_key_matching(
 
         // Get entity vertices. Padded with -1 if fewer than
         // max_vertices_per_entity
+        // NOTE Entity orientation is determined by vertex ordering. The
+        // orientation of an entity with respect to the cell may differ from its
+        // global mesh orientation. Hence, we reorder the vertices so that
+        // each entity's orientation agrees with their global orientation.
+        // FIXME This might be better below when the entity to vertex
+        // connectivity is computed
+        std::vector<std::int32_t> entity_vertices(ev.size());
+        for (std::size_t j = 0; j < ev.size(); ++j)
+          entity_vertices[j] = vertices[ev[j]];
+
+        // Orient the entities. Simply sort according to global vertex index
+        // for simplices
+        std::vector<std::int64_t> global_vertices(entity_vertices.size());
+        vertex_index_map.local_to_global(entity_vertices, global_vertices);
+
+        std::vector<std::size_t> perm(global_vertices.size());
+        std::iota(perm.begin(), perm.end(), 0);
+        std::ranges::sort(perm,
+                          [&global_vertices](std::size_t i0, std::size_t i1) {
+                            return global_vertices[i0] < global_vertices[i1];
+                          });
+        // For quadrilaterals, the vertex opposite the lowest vertex should
+        // be last
+        if (entity_type == mesh::CellType::quadrilateral)
+        {
+          std::size_t min_vertex_idx = perm[0];
+          std::size_t opposite_vertex_index = 3 - min_vertex_idx;
+          auto it = std::find(perm.begin(), perm.end(), opposite_vertex_index);
+          assert(it != perm.end());
+          std::rotate(it, it + 1, perm.end());
+        }
+
         for (std::size_t j = 0; j < ev.size(); ++j)
           entity_list[(cell_type_offsets[k] + idx) * num_vertices_per_entity
                       + j]
-              = vertices[ev[j]];
+              = entity_vertices[perm[j]];
       }
     }
   }
@@ -595,7 +617,7 @@ compute_entities_by_key_matching(
   for (std::size_t k = 0; k < cell_lists.size(); ++k)
   {
     auto cells = std::get<1>(cell_lists[k]);
-    const std::size_t num_cells = cells->num_nodes();
+    [[maybe_unused]] const std::size_t num_cells = cells->num_nodes();
     auto cell_map = std::get<2>(cell_lists[k]);
     int num_entities_per_cell = cell_type_entities[k].size();
     assert(cell_map->size_local() + cell_map->num_ghosts() == (int)num_cells);
@@ -726,7 +748,7 @@ compute_from_map(const graph::AdjacencyList<std::int32_t>& c_d0_0,
       auto v = vref->links(i);
       for (int j = 0; j < 2; ++j)
         key[j] = e0[v[j]];
-      std::sort(key.begin(), key.end());
+      std::ranges::sort(key);
       auto it = edge_to_index.find(key);
       assert(it != edge_to_index.end());
       connections.push_back(it->second);
@@ -745,7 +767,7 @@ std::tuple<std::vector<std::shared_ptr<graph::AdjacencyList<std::int32_t>>>,
 mesh::compute_entities(MPI_Comm comm, const Topology& topology, int dim,
                        int index)
 {
-  LOG(INFO) << "Computing mesh entities of dimension " << dim;
+  spdlog::info("Computing mesh entities of dimension {}", dim);
   const int tdim = topology.dim();
 
   // Vertices must always exist
@@ -782,13 +804,13 @@ mesh::compute_entities(MPI_Comm comm, const Topology& topology, int dim,
     cell_lists[i] = {cell_types[i], cells, cell_map};
   }
 
-  auto [d0, d1, im, interprocess_facets] = compute_entities_by_key_matching(
+  auto [d0, d1, im, interprocess_entities] = compute_entities_by_key_matching(
       comm, cell_lists, *vertex_map, entity_type, dim);
 
   return {d0,
           std::make_shared<graph::AdjacencyList<std::int32_t>>(std::move(d1)),
           std::make_shared<common::IndexMap>(std::move(im)),
-          std::move(interprocess_facets)};
+          std::move(interprocess_entities)};
 }
 //-----------------------------------------------------------------------------
 std::array<std::shared_ptr<graph::AdjacencyList<std::int32_t>>, 2>
@@ -796,9 +818,9 @@ mesh::compute_connectivity(const Topology& topology,
                            std::pair<std::int8_t, std::int8_t> d0,
                            std::pair<std::int8_t, std::int8_t> d1)
 {
-  LOG(INFO) << "Requesting connectivity (" << std::to_string(d0.first) << ","
-            << std::to_string(d0.second) << ") - (" << std::to_string(d1.first)
-            << "," << std::to_string(d1.second) << ")";
+  spdlog::info("Requesting connectivity ({}, {}) - ({}, {})",
+               std::to_string(d0.first), std::to_string(d0.second),
+               std::to_string(d1.first), std::to_string(d1.second));
 
   // Return if connectivity has already been computed
   if (topology.connectivity(d0, d1))
@@ -856,8 +878,8 @@ mesh::compute_connectivity(const Topology& topology,
       auto c_d1_d0 = std::make_shared<graph::AdjacencyList<std::int32_t>>(
           compute_from_map(*c_d1_0, *c_d0_0));
 
-      LOG(INFO) << "Computing mesh connectivity " << d0.first << " - "
-                << d1.first << " from transpose.";
+      spdlog::info("Computing mesh connectivity {}-{} from transpose.",
+                   d0.first, d1.first);
       auto c_d0_d1 = std::make_shared<graph::AdjacencyList<std::int32_t>>(
           compute_from_transpose(*c_d1_d0, c_d0_0->num_nodes()));
       return {c_d0_d1, c_d1_d0};
@@ -867,8 +889,8 @@ mesh::compute_connectivity(const Topology& topology,
       assert(c_d0_0);
       assert(topology.connectivity(d1, d0));
 
-      LOG(INFO) << "Computing mesh connectivity " << std::to_string(d0.first)
-                << " - " << std::to_string(d1.first) << " from transpose.";
+      spdlog::info("Computing mesh connectivity {}-{} from transpose.",
+                   std::to_string(d0.first), std::to_string(d1.first));
       auto c_d0_d1 = std::make_shared<graph::AdjacencyList<std::int32_t>>(
           compute_from_transpose(*topology.connectivity(d1, d0),
                                  c_d0_0->num_nodes()));
