@@ -9,6 +9,7 @@
 #include "Mesh.h"
 #include "Topology.h"
 #include "graphbuild.h"
+#include <algorithm>
 #include <basix/mdspan.hpp>
 #include <concepts>
 #include <dolfinx/graph/AdjacencyList.h>
@@ -54,7 +55,7 @@ void reorder_list(std::span<T> list, std::span<const std::int32_t> nodemap)
   {
     auto links_old = std::span(orig.data() + n * degree, degree);
     auto links_new = list.subspan(nodemap[n] * degree, degree);
-    std::copy(links_old.begin(), links_old.end(), links_new.begin());
+    std::ranges::copy(links_old, links_new.begin());
   }
 }
 
@@ -103,12 +104,17 @@ compute_vertex_coords_boundary(const mesh::Mesh<T>& mesh, int dim,
     }
 
     // Build vector of boundary vertices
-    std::sort(vertices.begin(), vertices.end());
-    vertices.erase(std::unique(vertices.begin(), vertices.end()),
-                   vertices.end());
-    std::sort(entities.begin(), entities.end());
-    entities.erase(std::unique(entities.begin(), entities.end()),
-                   entities.end());
+    {
+      std::ranges::sort(vertices);
+      auto [unique_end, range_end] = std::ranges::unique(vertices);
+      vertices.erase(unique_end, range_end);
+    }
+
+    {
+      std::ranges::sort(entities);
+      auto [unique_end, range_end] = std::ranges::unique(entities);
+      entities.erase(unique_end, range_end);
+    }
   }
 
   // Get geometry data
@@ -166,18 +172,19 @@ std::vector<std::int32_t> exterior_facet_indices(const Topology& topology);
 ///
 /// @param[in] comm MPI Communicator
 /// @param[in] nparts Number of partitions
-/// @param[in] cell_type Type of cell in mesh
-/// @param[in] cells Cells on this process. The ith entry in list
-/// contains the global indices for the cell vertices. Each cell can
-/// appear only once across all processes. The cell vertex indices are
-/// not necessarily contiguous globally, i.e. the maximum index across
-/// all processes can be greater than the number of vertices. High-order
-/// 'nodes', e.g. mid-side points, should not be included.
+/// @param[in] cell_types Cell types in the mesh
+/// @param[in] cells Lists of cells of each cell type. cells[i] is a flattened
+/// row major 2D array of shape (num_cells, num_cell_vertices) for cell_types[i]
+/// on this process, containing the global indices for the cell vertices. Each
+/// cell can appear only once across all processes. The cell vertex indices are
+/// not necessarily contiguous globally, i.e. the maximum index across all
+/// processes can be greater than the number of vertices. High-order 'nodes',
+/// e.g. mid-side points, should not be included.
 /// @return Destination ranks for each cell on this process
 /// @note Cells can have multiple destination ranks, when ghosted.
 using CellPartitionFunction = std::function<graph::AdjacencyList<std::int32_t>(
-    MPI_Comm comm, int nparts, CellType cell_type,
-    const graph::AdjacencyList<std::int64_t>& cells)>;
+    MPI_Comm comm, int nparts, const std::vector<CellType>& cell_types,
+    const std::vector<std::span<const std::int64_t>>& cells)>;
 
 /// @brief Extract topology from cell data, i.e. extract cell vertices.
 /// @param[in] cell_type The cell shape
@@ -272,14 +279,8 @@ std::vector<T> cell_normals(const Mesh<T>& mesh, int dim,
 
   // Find geometry nodes for topology entities
   std::span<const T> x = mesh.geometry().x();
-
-  // Orient cells if they are tetrahedron
-  bool orient = false;
-  if (topology->cell_type() == CellType::tetrahedron)
-    orient = true;
-
   std::vector<std::int32_t> geometry_entities
-      = entities_to_geometry(mesh, dim, entities, orient);
+      = entities_to_geometry(mesh, dim, entities, false);
 
   const std::size_t shape1 = geometry_entities.size() / entities.size();
   std::vector<T> n(entities.size() * 3);
@@ -299,8 +300,8 @@ std::vector<T> cell_normals(const Mesh<T>& mesh, int dim,
 
       // Define normal by rotating tangent counter-clockwise
       std::array<T, 3> t;
-      std::transform(p[1].begin(), p[1].end(), p[0].begin(), t.begin(),
-                     [](auto x, auto y) { return x - y; });
+      std::ranges::transform(p[1], p[0], t.begin(),
+                             [](auto x, auto y) { return x - y; });
 
       T norm = std::sqrt(t[0] * t[0] + t[1] * t[1]);
       std::span<T, 3> ni(n.data() + 3 * i, 3);
@@ -324,16 +325,16 @@ std::vector<T> cell_normals(const Mesh<T>& mesh, int dim,
 
       // Compute (p1 - p0) and (p2 - p0)
       std::array<T, 3> dp1, dp2;
-      std::transform(p[1].begin(), p[1].end(), p[0].begin(), dp1.begin(),
-                     [](auto x, auto y) { return x - y; });
-      std::transform(p[2].begin(), p[2].end(), p[0].begin(), dp2.begin(),
-                     [](auto x, auto y) { return x - y; });
+      std::ranges::transform(p[1], p[0], dp1.begin(),
+                             [](auto x, auto y) { return x - y; });
+      std::ranges::transform(p[2], p[0], dp2.begin(),
+                             [](auto x, auto y) { return x - y; });
 
       // Define cell normal via cross product of first two edges
       std::array<T, 3> ni = math::cross(dp1, dp2);
       T norm = std::sqrt(ni[0] * ni[0] + ni[1] * ni[1] + ni[2] * ni[2]);
-      std::transform(ni.begin(), ni.end(), std::next(n.begin(), 3 * i),
-                     [norm](auto x) { return x / norm; });
+      std::ranges::transform(ni, std::next(n.begin(), 3 * i),
+                             [norm](auto x) { return x / norm; });
     }
 
     return n;
@@ -353,16 +354,16 @@ std::vector<T> cell_normals(const Mesh<T>& mesh, int dim,
 
       // Compute (p1 - p0) and (p2 - p0)
       std::array<T, 3> dp1, dp2;
-      std::transform(p[1].begin(), p[1].end(), p[0].begin(), dp1.begin(),
-                     [](auto x, auto y) { return x - y; });
-      std::transform(p[2].begin(), p[2].end(), p[0].begin(), dp2.begin(),
-                     [](auto x, auto y) { return x - y; });
+      std::ranges::transform(p[1], p[0], dp1.begin(),
+                             [](auto x, auto y) { return x - y; });
+      std::ranges::transform(p[2], p[0], dp2.begin(),
+                             [](auto x, auto y) { return x - y; });
 
       // Define cell normal via cross product of first two edges
       std::array<T, 3> ni = math::cross(dp1, dp2);
       T norm = std::sqrt(ni[0] * ni[0] + ni[1] * ni[1] + ni[2] * ni[2]);
-      std::transform(ni.begin(), ni.end(), std::next(n.begin(), 3 * i),
-                     [norm](auto x) { return x / norm; });
+      std::ranges::transform(ni, std::next(n.begin(), 3 * i),
+                             [norm](auto x) { return x / norm; });
     }
 
     return n;
@@ -386,7 +387,6 @@ std::vector<T> compute_midpoints(const Mesh<T>& mesh, int dim,
   std::span<const T> x = mesh.geometry().x();
 
   // Build map from entity -> geometry dof
-  // FIXME: This assumes a linear geometry.
   const std::vector<std::int32_t> e_to_g
       = entities_to_geometry(mesh, dim, entities, false);
   std::size_t shape1 = e_to_g.size() / entities.size();
@@ -399,9 +399,9 @@ std::vector<T> compute_midpoints(const Mesh<T>& mesh, int dim,
     for (auto row : rows)
     {
       std::span<const T, 3> xg(x.data() + 3 * row, 3);
-      std::transform(p.begin(), p.end(), xg.begin(), p.begin(),
-                     [size = rows.size()](auto x, auto y)
-                     { return x + y / size; });
+      std::ranges::transform(p, xg, p.begin(),
+                             [size = rows.size()](auto x, auto y)
+                             { return x + y / size; });
     }
   }
 
@@ -610,125 +610,123 @@ std::vector<std::int32_t> locate_entities_boundary(const Mesh<T>& mesh, int dim,
   return entities;
 }
 
-/// @brief Determine the indices in the geometry data for each vertex of
-/// the given mesh entities.
-///
-/// @warning This function should not be used unless there is no
-/// alternative. It may be removed in the future.
+/// @brief Compute the geometry degrees of freedom associated with
+/// the closure of a given set of cell entities.
 ///
 /// @param[in] mesh The mesh.
 /// @param[in] dim Topological dimension of the entities of interest.
-/// @param[in] entities Entity indices (local) to compute the vertex
-/// geometry indices for.
-/// @param[in] orient If true, in 3D, reorients facets to have
-/// consistent normal direction.
-/// @return Indices in the geometry array for the entity vertices. The
-/// shape is `(num_entities, num_vertices_per_entity)` and the storage
-/// is row-major. The index `indices[i, j]` is the position in the
-/// geometry array of the `j`-th vertex of the `entity[i]`.
+/// @param[in] entities Entity indices (local to process).
+/// @param[in] permute If `true`, permute the DOFs such that they are
+/// consistent with the orientation of `dim`-dimensional mesh entities.
+/// This requires `create_entity_permutations` to be called first.
+/// @return The geometry DOFs associated with the closure of each entity
+/// in `entities`. The shape is `(num_entities, num_xdofs_per_entity)`
+/// and the storage is row-major. The index `indices[i, j]` is the
+/// position in the geometry array of the `j`-th vertex of the
+/// `entity[i]`.
+///
+/// @pre The mesh connectivities `dim -> mesh.topology().dim()` and
+/// `mesh.topology().dim() -> dim` must have been computed. Otherwise an
+/// exception is thrown.
 template <std::floating_point T>
 std::vector<std::int32_t>
 entities_to_geometry(const Mesh<T>& mesh, int dim,
-                     std::span<const std::int32_t> entities, bool orient)
+                     std::span<const std::int32_t> entities,
+                     bool permute = false)
 {
   auto topology = mesh.topology();
   assert(topology);
-
   CellType cell_type = topology->cell_type();
   if (cell_type == CellType::prism and dim == 2)
     throw std::runtime_error("More work needed for prism cells");
-  if (orient and (cell_type != CellType::tetrahedron or dim != 2))
-    throw std::runtime_error("Can only orient facets of a tetrahedral mesh");
-
-  const Geometry<T>& geometry = mesh.geometry();
-  auto x = geometry.x();
 
   const int tdim = topology->dim();
-  mesh.topology_mutable()->create_entities(dim);
-  mesh.topology_mutable()->create_connectivity(dim, tdim);
-  mesh.topology_mutable()->create_connectivity(dim, 0);
-  mesh.topology_mutable()->create_connectivity(tdim, 0);
-
+  const Geometry<T>& geometry = mesh.geometry();
   auto xdofs = geometry.dofmap();
+
+  // Get the DOF layout and the number of DOFs per entity
+  const fem::CoordinateElement<T>& coord_ele = geometry.cmap();
+  const fem::ElementDofLayout layout = coord_ele.create_dof_layout();
+  const std::size_t num_entity_dofs = layout.num_entity_closure_dofs(dim);
+  std::vector<std::int32_t> entity_xdofs;
+  entity_xdofs.reserve(entities.size() * num_entity_dofs);
+
+  // Get the element's closure DOFs
+  const std::vector<std::vector<std::vector<int>>>& closure_dofs_all
+      = layout.entity_closure_dofs_all();
+
+  // Special case when dim == tdim (cells)
+  if (dim == tdim)
+  {
+    for (std::size_t i = 0; i < entities.size(); ++i)
+    {
+      const std::int32_t c = entities[i];
+      // Extract degrees of freedom
+      auto x_c = MDSPAN_IMPL_STANDARD_NAMESPACE::submdspan(
+          xdofs, c, MDSPAN_IMPL_STANDARD_NAMESPACE::full_extent);
+      for (std::int32_t entity_dof : closure_dofs_all[tdim][0])
+        entity_xdofs.push_back(x_c[entity_dof]);
+    }
+    return entity_xdofs;
+  }
+
+  assert(dim != tdim);
+
   auto e_to_c = topology->connectivity(dim, tdim);
   if (!e_to_c)
   {
     throw std::runtime_error(
-        "Entity-to-cell connectivity has not been computed.");
+        "Entity-to-cell connectivity has not been computed. Missing dims "
+        + std::to_string(dim) + "->" + std::to_string(tdim));
   }
 
-  auto e_to_v = topology->connectivity(dim, 0);
-  if (!e_to_v)
+  auto c_to_e = topology->connectivity(tdim, dim);
+  if (!c_to_e)
   {
     throw std::runtime_error(
-        "Entity-to-vertex connectivity has not been computed.");
+        "Cell-to-entity connectivity has not been computed. Missing dims "
+        + std::to_string(tdim) + "->" + std::to_string(dim));
   }
 
-  auto c_to_v = topology->connectivity(tdim, 0);
-  if (!e_to_v)
-  {
-    throw std::runtime_error(
-        "Cell-to-vertex connectivity has not been computed.");
-  }
+  // Get the cell info, which is needed to permute the closure dofs
+  std::span<const std::uint32_t> cell_info;
+  if (permute)
+    cell_info = std::span(mesh.topology()->get_cell_permutation_info());
 
-  const std::size_t num_vertices
-      = num_cell_vertices(cell_entity_type(cell_type, dim, 0));
-  std::vector<std::int32_t> geometry_idx(entities.size() * num_vertices);
   for (std::size_t i = 0; i < entities.size(); ++i)
   {
-    const std::int32_t idx = entities[i];
-    // Always pick the second cell to be consistent with the e_to_v connectivity
-    const std::int32_t cell = e_to_c->links(idx).back();
-    auto ev = e_to_v->links(idx);
-    assert(ev.size() == num_vertices);
-    auto cv = c_to_v->links(cell);
-    std::span<const std::int32_t> xc(
-        xdofs.data_handle() + xdofs.extent(1) * cell, xdofs.extent(1));
-    for (std::size_t j = 0; j < num_vertices; ++j)
+    const std::int32_t e = entities[i];
+
+    // Get a cell connected to the entity
+    assert(!e_to_c->links(e).empty());
+    std::int32_t c = e_to_c->links(e).front();
+
+    // Get the local index of the entity
+    std::span<const std::int32_t> cell_entities = c_to_e->links(c);
+    auto it = std::find(cell_entities.begin(), cell_entities.end(), e);
+    assert(it != cell_entities.end());
+    std::size_t local_entity = std::distance(cell_entities.begin(), it);
+
+    std::vector<std::int32_t> closure_dofs(closure_dofs_all[dim][local_entity]);
+
+    // Cell sub-entities must be permuted so that their local orientation agrees
+    // with their global orientation
+    if (permute)
     {
-      int k = std::distance(cv.begin(), std::find(cv.begin(), cv.end(), ev[j]));
-      assert(k < (int)cv.size());
-      geometry_idx[i * num_vertices + j] = xc[k];
+      mesh::CellType entity_type
+          = mesh::cell_entity_type(cell_type, dim, local_entity);
+      coord_ele.permute_subentity_closure(closure_dofs, cell_info[c],
+                                          entity_type, local_entity);
     }
 
-    if (orient)
-    {
-      // Compute cell midpoint
-      std::array<T, 3> midpoint = {0, 0, 0};
-      for (std::int32_t j : xc)
-        for (int k = 0; k < 3; ++k)
-          midpoint[k] += x[3 * j + k];
-      std::transform(midpoint.begin(), midpoint.end(), midpoint.begin(),
-                     [size = xc.size()](auto x) { return x / size; });
-
-      // Compute vector triple product of two edges and vector to midpoint
-      std::array<T, 3> p0, p1, p2;
-      std::copy_n(std::next(x.begin(), 3 * geometry_idx[i * num_vertices + 0]),
-                  3, p0.begin());
-      std::copy_n(std::next(x.begin(), 3 * geometry_idx[i * num_vertices + 1]),
-                  3, p1.begin());
-      std::copy_n(std::next(x.begin(), 3 * geometry_idx[i * num_vertices + 2]),
-                  3, p2.begin());
-
-      std::array<T, 9> a;
-      std::transform(midpoint.begin(), midpoint.end(), p0.begin(), a.begin(),
-                     [](auto x, auto y) { return x - y; });
-      std::transform(p1.begin(), p1.end(), p0.begin(), std::next(a.begin(), 3),
-                     [](auto x, auto y) { return x - y; });
-      std::transform(p2.begin(), p2.end(), p0.begin(), std::next(a.begin(), 6),
-                     [](auto x, auto y) { return x - y; });
-
-      // Midpoint direction should be opposite to normal, hence this
-      // should be negative. Switch points if not.
-      if (math::det(a.data(), {3, 3}) > 0.0)
-      {
-        std::swap(geometry_idx[i * num_vertices + 1],
-                  geometry_idx[i * num_vertices + 2]);
-      }
-    }
+    // Extract degrees of freedom
+    auto x_c = MDSPAN_IMPL_STANDARD_NAMESPACE::submdspan(
+        xdofs, c, MDSPAN_IMPL_STANDARD_NAMESPACE::full_extent);
+    for (std::int32_t entity_dof : closure_dofs)
+      entity_xdofs.push_back(x_c[entity_dof]);
   }
 
-  return geometry_idx;
+  return entity_xdofs;
 }
 
 /// Create a function that computes destination rank for mesh cells in
@@ -772,7 +770,7 @@ compute_incident_entities(const Topology& topology,
 /// 'nodes' will be included. See dolfinx::io::cells for examples of the
 /// Basix ordering.
 /// @param[in] element Coordinate element for the cells.
-/// @param[in] commg
+/// @param[in] commg Communicator for geometry
 /// @param[in] x Geometry data ('node' coordinates). Row-major storage.
 /// The global index of the `i`th node (row) in `x` is taken as `i` plus
 /// the process offset  on`comm`, The offset  is the sum of `x` rows on
@@ -793,7 +791,7 @@ Mesh<typename std::remove_reference_t<typename U::value_type>> create_mesh(
   const fem::ElementDofLayout doflayout = element.create_dof_layout();
 
   const int num_cell_vertices = mesh::num_cell_vertices(element.cell_shape());
-  const int num_cell_nodes = doflayout.num_dofs();
+  std::size_t num_cell_nodes = doflayout.num_dofs();
 
   // Note: `extract_topology` extracts topology data, i.e. just the
   // vertices. For P1 geometry this should just be the identity
@@ -803,44 +801,47 @@ Mesh<typename std::remove_reference_t<typename U::value_type>> create_mesh(
   // `extract_topology` could be skipped for 'P1 geometry' elements
 
   // -- Partition topology across ranks of comm
-  graph::AdjacencyList<std::int64_t> cells1(0);
-  // std::vector<std::int64_t> cells1;
+  std::vector<std::int64_t> cells1;
   std::vector<std::int64_t> original_idx1;
   std::vector<int> ghost_owners;
   if (partitioner)
   {
+    spdlog::info("Using partitioner with {} cell data", cells.size());
     graph::AdjacencyList<std::int32_t> dest(0);
     if (commt != MPI_COMM_NULL)
     {
       int size = dolfinx::MPI::size(comm);
-      auto t = graph::regular_adjacency_list(
-          extract_topology(element.cell_shape(), doflayout, cells),
-          num_cell_vertices);
-      dest = partitioner(commt, size, celltype, t);
+      auto t = extract_topology(element.cell_shape(), doflayout, cells);
+      dest = partitioner(commt, size, {celltype}, {t});
     }
 
     // Distribute cells (topology, includes higher-order 'nodes') to
     // destination rank
-    std::vector<int> src;
-    auto _cells = graph::regular_adjacency_list(
-        std::vector(cells.begin(), cells.end()), num_cell_nodes);
-    std::tie(cells1, src, original_idx1, ghost_owners)
-        = graph::build::distribute(comm, _cells, dest);
+    assert(cells.size() % num_cell_nodes == 0);
+    std::size_t num_cells = cells.size() / num_cell_nodes;
+    std::vector<int> src_ranks;
+    std::tie(cells1, src_ranks, original_idx1, ghost_owners)
+        = graph::build::distribute(comm, cells, {num_cells, num_cell_nodes},
+                                   dest);
+    spdlog::debug("Got {} cells from distribution", cells1.size());
   }
   else
   {
-    cells1 = graph::regular_adjacency_list(
-        std::vector(cells.begin(), cells.end()), num_cell_nodes);
-    std::int64_t offset(0), num_owned(cells1.num_nodes());
+    cells1 = std::vector<std::int64_t>(cells.begin(), cells.end());
+    assert(cells1.size() % num_cell_nodes == 0);
+    std::int64_t offset = 0;
+    std::int64_t num_owned = cells1.size() / num_cell_nodes;
     MPI_Exscan(&num_owned, &offset, 1, MPI_INT64_T, MPI_SUM, comm);
-    original_idx1.resize(cells1.num_nodes());
+    original_idx1.resize(num_owned);
     std::iota(original_idx1.begin(), original_idx1.end(), offset);
   }
 
   // Extract cell 'topology', i.e. extract the vertices for each cell
   // and discard any 'higher-order' nodes
   std::vector<std::int64_t> cells1_v
-      = extract_topology(celltype, doflayout, cells1.array());
+      = extract_topology(celltype, doflayout, cells1);
+  spdlog::info("Extract basic topology: {}->{}", cells1.size(),
+               cells1_v.size());
 
   // Build local dual graph for owned cells to (i) get list of vertices
   // on the process boundary and (ii) apply re-ordering to cells for
@@ -852,10 +853,11 @@ Mesh<typename std::remove_reference_t<typename U::value_type>> create_mesh(
     std::vector<std::int32_t> cell_offsets(num_owned_cells + 1, 0);
     for (std::size_t i = 1; i < cell_offsets.size(); ++i)
       cell_offsets[i] = cell_offsets[i - 1] + num_cell_vertices;
+    spdlog::info("Build local dual graph");
     auto [graph, unmatched_facets, max_v, facet_attached_cells]
         = build_local_dual_graph(
-            celltype,
-            std::span(cells1_v.data(), num_owned_cells * num_cell_vertices));
+            std::vector{celltype},
+            {std::span(cells1_v.data(), num_owned_cells * num_cell_vertices)});
     const std::vector<int> remap = graph::reorder_gps(graph);
 
     // Create re-ordered cell lists (leaves ghosts unchanged)
@@ -867,15 +869,15 @@ Mesh<typename std::remove_reference_t<typename U::value_type>> create_mesh(
                 std::next(_original_idx.begin(), num_owned_cells));
     impl::reorder_list(
         std::span(cells1_v.data(), remap.size() * num_cell_vertices), remap);
-    impl::reorder_list(
-        std::span(cells1.array().data(), remap.size() * num_cell_nodes), remap);
+    impl::reorder_list(std::span(cells1.data(), remap.size() * num_cell_nodes),
+                       remap);
     original_idx1 = _original_idx;
 
     // Boundary vertices are marked as 'unknown'
     boundary_v = unmatched_facets;
-    std::sort(boundary_v.begin(), boundary_v.end());
-    boundary_v.erase(std::unique(boundary_v.begin(), boundary_v.end()),
-                     boundary_v.end());
+    std::ranges::sort(boundary_v);
+    auto [unique_end, range_end] = std::ranges::unique(boundary_v);
+    boundary_v.erase(unique_end, range_end);
 
     // Remove -1 if it occurs in boundary vertices (may occur in mixed
     // topology)
@@ -897,15 +899,246 @@ Mesh<typename std::remove_reference_t<typename U::value_type>> create_mesh(
 
   // Build list of unique (global) node indices from cells1 and
   // distribute coordinate data
-  std::vector<std::int64_t> nodes1 = cells1.array();
-  dolfinx::radix_sort(std::span(nodes1));
-  nodes1.erase(std::unique(nodes1.begin(), nodes1.end()), nodes1.end());
+  std::vector<std::int64_t> nodes1 = cells1;
+  dolfinx::radix_sort(nodes1);
+  auto [unique_end, range_end] = std::ranges::unique(nodes1);
+  nodes1.erase(unique_end, range_end);
+
   std::vector coords
       = dolfinx::MPI::distribute_data(comm, nodes1, commg, x, xshape[1]);
 
   // Create geometry object
-  Geometry geometry = create_geometry(topology, element, nodes1, cells1.array(),
-                                      coords, xshape[1]);
+  Geometry geometry
+      = create_geometry(topology, element, nodes1, cells1, coords, xshape[1]);
+
+  return Mesh(comm, std::make_shared<Topology>(std::move(topology)),
+              std::move(geometry));
+}
+
+/// @brief Create a distributed mixed-topology mesh from mesh data using a
+/// provided graph partitioning function for determining the parallel
+/// distribution of the mesh.
+///
+/// From mesh input data that is distributed across processes, a
+/// distributed mesh::Mesh is created. If the partitioning function is
+/// not callable, i.e. it does not store a callable function, no
+/// re-distribution of cells is done.
+///
+/// @note This is an experimental specialised version of `create_mesh` for mixed
+/// topology meshes, and does not include cell reordering.
+///
+/// @param[in] comm Communicator to build the mesh on.
+/// @param[in] commt Communicator that the topology data (`cells`) is
+/// distributed on. This should be `MPI_COMM_NULL` for ranks that should
+/// not participate in computing the topology partitioning.
+/// @param[in] cells Cells on the calling process, as a list of lists,
+/// one list for each cell type (or an empty list if there are no cells of that
+/// type on this process). The cells are defined by their 'nodes' (using global
+/// indices) following the Basix ordering, and concatenated to form a flattened
+/// list.  For lowest order cells this will be just the cell vertices. For
+/// higher-order cells, other cells 'nodes' will be included. See
+/// dolfinx::io::cells for examples of the Basix ordering.
+/// @param[in] elements Coordinate elements for the cells, one for each cell
+/// type in the mesh. In parallel, these must be the same on all processes.
+/// @param[in] commg Communicator for geometry
+/// @param[in] x Geometry data ('node' coordinates). Row-major storage.
+/// The global index of the `i`th node (row) in `x` is taken as `i` plus
+/// the process offset  on`comm`, The offset  is the sum of `x` rows on
+/// all processed with a lower rank than the caller.
+/// @param[in] xshape Shape of the `x` data.
+/// @param[in] partitioner Graph partitioner that computes the owning
+/// rank for each cell. If not callable, cells are not redistributed.
+/// @return A mesh distributed on the communicator `comm`.
+template <typename U>
+Mesh<typename std::remove_reference_t<typename U::value_type>> create_mesh(
+    MPI_Comm comm, MPI_Comm commt,
+    const std::vector<std::span<const std::int64_t>>& cells,
+    const std::vector<fem::CoordinateElement<
+        typename std::remove_reference_t<typename U::value_type>>>& elements,
+    MPI_Comm commg, const U& x, std::array<std::size_t, 2> xshape,
+    const CellPartitionFunction& partitioner)
+{
+  assert(cells.size() == elements.size());
+  std::int32_t num_cell_types = cells.size();
+  std::vector<CellType> celltypes;
+  std::ranges::transform(elements, std::back_inserter(celltypes),
+                         [](auto e) { return e.cell_shape(); });
+  std::vector<fem::ElementDofLayout> doflayouts;
+  std::ranges::transform(elements, std::back_inserter(doflayouts),
+                         [](auto e) { return e.create_dof_layout(); });
+
+  // Note: `extract_topology` extracts topology data, i.e. just the
+  // vertices. For P1 geometry this should just be the identity
+  // operator. For other elements the filtered lists may have 'gaps',
+  // i.e. the indices might not be contiguous.
+  //
+  // `extract_topology` could be skipped for 'P1 geometry' elements
+
+  // -- Partition topology across ranks of comm
+  std::vector<std::vector<std::int64_t>> cells1(num_cell_types);
+  std::vector<std::vector<std::int64_t>> original_idx1(num_cell_types);
+  std::vector<std::vector<int>> ghost_owners(num_cell_types);
+  if (partitioner)
+  {
+    spdlog::info("Using partitioner with cell data ({} cell types)",
+                 num_cell_types);
+    graph::AdjacencyList<std::int32_t> dest(0);
+    if (commt != MPI_COMM_NULL)
+    {
+      int size = dolfinx::MPI::size(comm);
+      std::vector<std::vector<std::int64_t>> t(num_cell_types);
+      std::vector<std::span<const std::int64_t>> tspan(num_cell_types);
+      for (std::int32_t i = 0; i < num_cell_types; ++i)
+      {
+        t[i] = extract_topology(celltypes[i], doflayouts[i], cells[i]);
+        tspan[i] = std::span(t[i]);
+      }
+      dest = partitioner(commt, size, celltypes, tspan);
+    }
+
+    std::int32_t cell_offset = 0;
+    for (std::int32_t i = 0; i < num_cell_types; ++i)
+    {
+      std::size_t num_cell_nodes = doflayouts[i].num_dofs();
+      assert(cells[i].size() % num_cell_nodes == 0);
+      std::size_t num_cells = cells[i].size() / num_cell_nodes;
+
+      // Extract destination AdjacencyList for this cell type
+      std::vector<std::int32_t> offsets_i(
+          std::next(dest.offsets().begin(), cell_offset),
+          std::next(dest.offsets().begin(), cell_offset + num_cells + 1));
+      std::vector<std::int32_t> data_i(
+          std::next(dest.array().begin(), offsets_i.front()),
+          std::next(dest.array().begin(), offsets_i.back()));
+      std::int32_t offset_0 = offsets_i.front();
+      std::ranges::for_each(offsets_i,
+                            [&offset_0](std::int32_t& j) { j -= offset_0; });
+      graph::AdjacencyList<std::int32_t> dest_i(data_i, offsets_i);
+      cell_offset += num_cells;
+
+      // Distribute cells (topology, includes higher-order 'nodes') to
+      // destination rank
+      std::vector<int> src_ranks;
+      std::tie(cells1[i], src_ranks, original_idx1[i], ghost_owners[i])
+          = graph::build::distribute(comm, cells[i],
+                                     {num_cells, num_cell_nodes}, dest_i);
+      spdlog::debug("Got {} cells from distribution", cells1[i].size());
+    }
+  }
+  else
+  {
+    // No partitioning, construct a global index
+    std::int64_t num_owned = 0;
+    for (std::int32_t i = 0; i < num_cell_types; ++i)
+    {
+      cells1[i] = std::vector<std::int64_t>(cells[i].begin(), cells[i].end());
+      std::int32_t num_cell_nodes = doflayouts[i].num_dofs();
+      assert(cells1[i].size() % num_cell_nodes == 0);
+      original_idx1[i].resize(cells1[i].size() / num_cell_nodes);
+      num_owned += original_idx1[i].size();
+    }
+    // Add on global offset
+    std::int64_t global_offset = 0;
+    MPI_Exscan(&num_owned, &global_offset, 1, MPI_INT64_T, MPI_SUM, comm);
+    for (std::int32_t i = 0; i < num_cell_types; ++i)
+    {
+      std::iota(original_idx1[i].begin(), original_idx1[i].end(),
+                global_offset);
+      global_offset += original_idx1[i].size();
+    }
+  }
+
+  // Extract cell 'topology', i.e. extract the vertices for each cell
+  // and discard any 'higher-order' nodes
+  std::vector<std::vector<std::int64_t>> cells1_v(num_cell_types);
+  for (std::int32_t i = 0; i < num_cell_types; ++i)
+  {
+    cells1_v[i] = extract_topology(celltypes[i], doflayouts[i], cells1[i]);
+    spdlog::info("Extract basic topology: {}->{}", cells1[i].size(),
+                 cells1_v[i].size());
+  }
+
+  // Build local dual graph for owned cells to (i) get list of vertices
+  // on the process boundary and (ii) apply re-ordering to cells for
+  // locality
+  std::vector<std::int64_t> boundary_v;
+  {
+    std::vector<std::span<const std::int64_t>> cells1_v_local_cells;
+
+    for (std::int32_t i = 0; i < num_cell_types; ++i)
+    {
+      std::int32_t num_cell_vertices = mesh::num_cell_vertices(celltypes[i]);
+      std::int32_t num_owned_cells
+          = cells1_v[i].size() / num_cell_vertices - ghost_owners[i].size();
+      cells1_v_local_cells.push_back(
+          std::span(cells1_v[i].data(), num_owned_cells * num_cell_vertices));
+    }
+    spdlog::info("Build local dual graph");
+    auto [graph, unmatched_facets, max_v, facet_attached_cells]
+        = build_local_dual_graph(celltypes, cells1_v_local_cells);
+
+    // TODO: in the original create_mesh(), cell reordering is done here.
+    // This needs reworking for mixed cell topology
+
+    // Boundary vertices are marked as 'unknown'
+    boundary_v = unmatched_facets;
+    std::ranges::sort(boundary_v);
+    auto [unique_end, range_end] = std::ranges::unique(boundary_v);
+    boundary_v.erase(unique_end, range_end);
+
+    // Remove -1 if it occurs in boundary vertices (may occur in mixed
+    // topology)
+    if (!boundary_v.empty() > 0 and boundary_v[0] == -1)
+      boundary_v.erase(boundary_v.begin());
+  }
+
+  spdlog::debug("Got {} boundary vertices", boundary_v.size());
+
+  // Create Topology
+
+  std::vector<std::span<const std::int64_t>> cells1_v_span;
+  std::ranges::transform(cells1_v, std::back_inserter(cells1_v_span),
+                         [](auto& c) { return std::span(c); });
+  std::vector<std::span<const std::int64_t>> original_idx1_span;
+  std::ranges::transform(original_idx1, std::back_inserter(original_idx1_span),
+                         [](auto& c) { return std::span(c); });
+  std::vector<std::span<const int>> ghost_owners_span;
+  std::ranges::transform(ghost_owners, std::back_inserter(ghost_owners_span),
+                         [](auto& c) { return std::span(c); });
+
+  Topology topology
+      = create_topology(comm, celltypes, cells1_v_span, original_idx1_span,
+                        ghost_owners_span, boundary_v);
+
+  // Create connectivities required higher-order geometries for creating
+  // a Geometry object
+  for (int i = 0; i < num_cell_types; ++i)
+  {
+    for (int e = 1; e < topology.dim(); ++e)
+      if (doflayouts[i].num_entity_dofs(e) > 0)
+        topology.create_entities(e);
+    if (elements[i].needs_dof_permutations())
+      topology.create_entity_permutations();
+  }
+
+  // Build list of unique (global) node indices from cells1 and
+  // distribute coordinate data
+  std::vector<std::int64_t> nodes1, nodes2;
+  for (std::vector<std::int64_t>& c : cells1)
+    nodes1.insert(nodes1.end(), c.begin(), c.end());
+  for (std::vector<std::int64_t>& c : cells1)
+    nodes2.insert(nodes2.end(), c.begin(), c.end());
+
+  dolfinx::radix_sort(nodes1);
+  auto [unique_end, range_end] = std::ranges::unique(nodes1);
+  nodes1.erase(unique_end, range_end);
+
+  std::vector coords
+      = dolfinx::MPI::distribute_data(comm, nodes1, commg, x, xshape[1]);
+
+  // Create geometry object
+  Geometry geometry
+      = create_geometry(topology, elements, nodes1, nodes2, coords, xshape[1]);
 
   return Mesh(comm, std::make_shared<Topology>(std::move(topology)),
               std::move(geometry));
@@ -945,6 +1178,99 @@ create_mesh(MPI_Comm comm, std::span<const std::int64_t> cells,
   }
 }
 
+/// @brief Create a sub-geometry from a mesh and a subset of mesh entities to
+/// be included. A sub-geometry is simply a `Geometry` object containing only
+/// the geometric information for the subset of entities. The entities may
+/// differ in topological dimension from the original mesh.
+///
+/// @param[in] mesh The full mesh.
+/// @param[in] dim Topological dimension of the sub-topology.
+/// @param[in] subentity_to_entity Map from sub-topology entity to the
+/// entity in the parent topology.
+/// @return A sub-geometry and a map from sub-geometry coordinate
+/// degree-of-freedom to the coordinate degree-of-freedom in `geometry`.
+template <std::floating_point T>
+std::pair<Geometry<T>, std::vector<int32_t>>
+create_subgeometry(const Mesh<T>& mesh, int dim,
+                   std::span<const std::int32_t> subentity_to_entity)
+{
+  const Geometry<T>& geometry = mesh.geometry();
+
+  // Get the geometry dofs in the sub-geometry based on the entities in
+  // sub-geometry
+  const fem::ElementDofLayout layout = geometry.cmap().create_dof_layout();
+
+  std::vector<std::int32_t> x_indices
+      = entities_to_geometry(mesh, dim, subentity_to_entity, true);
+
+  std::vector<std::int32_t> sub_x_dofs = x_indices;
+  std::ranges::sort(sub_x_dofs);
+  auto [unique_end, range_end] = std::ranges::unique(sub_x_dofs);
+  sub_x_dofs.erase(unique_end, range_end);
+
+  // Get the sub-geometry dofs owned by this process
+  auto x_index_map = geometry.index_map();
+  assert(x_index_map);
+
+  std::shared_ptr<common::IndexMap> sub_x_dof_index_map;
+  std::vector<std::int32_t> subx_to_x_dofmap;
+  {
+    auto [map, new_to_old] = common::create_sub_index_map(
+        *x_index_map, sub_x_dofs, common::IndexMapOrder::any, true);
+    sub_x_dof_index_map = std::make_shared<common::IndexMap>(std::move(map));
+    subx_to_x_dofmap = std::move(new_to_old);
+  }
+
+  // Create sub-geometry coordinates
+  std::span<const T> x = geometry.x();
+  std::int32_t sub_num_x_dofs = subx_to_x_dofmap.size();
+  std::vector<T> sub_x(3 * sub_num_x_dofs);
+  for (int i = 0; i < sub_num_x_dofs; ++i)
+  {
+    std::copy_n(std::next(x.begin(), 3 * subx_to_x_dofmap[i]), 3,
+                std::next(sub_x.begin(), 3 * i));
+  }
+
+  // Create geometry to sub-geometry  map
+  std::vector<std::int32_t> x_to_subx_dof_map(
+      x_index_map->size_local() + x_index_map->num_ghosts(), -1);
+  for (std::size_t i = 0; i < subx_to_x_dofmap.size(); ++i)
+    x_to_subx_dof_map[subx_to_x_dofmap[i]] = i;
+
+  // Create sub-geometry dofmap
+  std::vector<std::int32_t> sub_x_dofmap;
+  sub_x_dofmap.reserve(x_indices.size());
+  std::ranges::transform(x_indices, std::back_inserter(sub_x_dofmap),
+                         [&x_to_subx_dof_map](auto x_dof)
+                         {
+                           assert(x_to_subx_dof_map[x_dof] != -1);
+                           return x_to_subx_dof_map[x_dof];
+                         });
+
+  // Create sub-geometry coordinate element
+  CellType sub_coord_cell
+      = cell_entity_type(geometry.cmap().cell_shape(), dim, 0);
+  // Special handling if point meshes, as they only support constant basis
+  // functions
+  int degree = geometry.cmap().degree();
+  if (sub_coord_cell == CellType::point)
+    degree = 0;
+  fem::CoordinateElement<T> sub_cmap(sub_coord_cell, degree,
+                                     geometry.cmap().variant());
+
+  // Sub-geometry input_global_indices
+  const std::vector<std::int64_t>& igi = geometry.input_global_indices();
+  std::vector<std::int64_t> sub_igi;
+  sub_igi.reserve(subx_to_x_dofmap.size());
+  std::ranges::transform(subx_to_x_dofmap, std::back_inserter(sub_igi),
+                         [&igi](auto sub_x_dof) { return igi[sub_x_dof]; });
+
+  // Create geometry
+  return {Geometry(sub_x_dof_index_map, std::move(sub_x_dofmap), {sub_cmap},
+                   std::move(sub_x), geometry.dim(), std::move(sub_igi)),
+          std::move(subx_to_x_dofmap)};
+}
+
 /// @brief Create a new mesh consisting of a subset of entities in a
 /// mesh.
 /// @param[in] mesh The mesh
@@ -969,8 +1295,9 @@ create_submesh(const Mesh<T>& mesh, int dim,
   mesh.topology_mutable()->create_entities(dim);
   mesh.topology_mutable()->create_connectivity(dim, tdim);
   mesh.topology_mutable()->create_connectivity(tdim, dim);
-  auto [geometry, subx_to_x_dofmap] = mesh::create_subgeometry(
-      *mesh.topology(), mesh.geometry(), dim, subentity_to_entity);
+  mesh.topology_mutable()->create_entity_permutations();
+  auto [geometry, subx_to_x_dofmap]
+      = mesh::create_subgeometry(mesh, dim, subentity_to_entity);
 
   return {Mesh(mesh.comm(), std::make_shared<Topology>(std::move(topology)),
                std::move(geometry)),
