@@ -4,11 +4,11 @@
 //
 // SPDX-License-Identifier:    LGPL-3.0-or-later
 
+#include "mesh.h"
 #include "MPICommWrapper.h"
 #include "array.h"
 #include "caster_mpi.h"
 #include "numpy_dtype.h"
-#include <cfloat>
 #include <dolfinx/common/IndexMap.h>
 #include <dolfinx/fem/CoordinateElement.h>
 #include <dolfinx/fem/ElementDofLayout.h>
@@ -21,7 +21,6 @@
 #include <dolfinx/mesh/graphbuild.h>
 #include <dolfinx/mesh/topologycomputation.h>
 #include <dolfinx/mesh/utils.h>
-#include <iostream>
 #include <memory>
 #include <nanobind/nanobind.h>
 #include <nanobind/ndarray.h>
@@ -36,63 +35,36 @@
 
 namespace nb = nanobind;
 
-namespace
+namespace dolfinx_wrappers
 {
-/// Wrap a Python graph partitioning function as a C++ function
-template <typename Functor>
-auto create_partitioner_cpp(Functor p)
+namespace part::impl
 {
-  return [p](MPI_Comm comm, int nparts,
-             const dolfinx::graph::AdjacencyList<std::int64_t>& local_graph,
-             bool ghosting)
-  {
-    return p(dolfinx_wrappers::MPICommWrapper(comm), nparts, local_graph,
-             ghosting);
-  };
-}
-
-/// Wrap a C++ cell partitioning function as a Python function
-template <typename Functor>
-auto create_cell_partitioner_py(Functor p)
-{
-  return [p](dolfinx_wrappers::MPICommWrapper comm, int n,
-             dolfinx::mesh::CellType cell_type,
-             const dolfinx::graph::AdjacencyList<std::int64_t>& cells)
-  { return p(comm.get(), n, cell_type, cells); };
-}
-
-using PythonPartitioningFunction
-    = std::function<dolfinx::graph::AdjacencyList<std::int32_t>(
-        dolfinx_wrappers::MPICommWrapper, int, dolfinx::mesh::CellType,
-        const dolfinx::graph::AdjacencyList<std::int64_t>&)>;
-
-using PythonCellPartitionFunction
-    = std::function<dolfinx::graph::AdjacencyList<std::int32_t>(
-        dolfinx_wrappers::MPICommWrapper, int, dolfinx::mesh::CellType,
-        const dolfinx::graph::AdjacencyList<std::int64_t>&)>;
-
-using CppCellPartitionFunction
-    = std::function<dolfinx::graph::AdjacencyList<std::int32_t>(
-        MPI_Comm, int, dolfinx::mesh::CellType,
-        const dolfinx::graph::AdjacencyList<std::int64_t>&)>;
-
-/// Wrap a Python cell graph partitioning function as a C++ function
 CppCellPartitionFunction
 create_cell_partitioner_cpp(const PythonCellPartitionFunction& p)
 {
   if (p)
   {
-    return [p](MPI_Comm comm, int n, dolfinx::mesh::CellType cell_type,
-               const dolfinx::graph::AdjacencyList<std::int64_t>& cells)
-    { return p(dolfinx_wrappers::MPICommWrapper(comm), n, cell_type, cells); };
+    return [p](MPI_Comm comm, int n,
+               const std::vector<dolfinx::mesh::CellType>& cell_types,
+               const std::vector<std::span<const std::int64_t>>& cells)
+    {
+      std::vector<nb::ndarray<const std::int64_t, nb::numpy>> cells_nb;
+      std::ranges::transform(
+          cells, std::back_inserter(cells_nb),
+          [](auto c)
+          {
+            return nb::ndarray<const std::int64_t, nb::numpy>(
+                c.data(), {c.size()}, nb::handle());
+          });
+
+      return p(dolfinx_wrappers::MPICommWrapper(comm), n, cell_types, cells_nb);
+    };
   }
   else
     return nullptr;
 }
-} // namespace
+} // namespace part::impl
 
-namespace dolfinx_wrappers
-{
 template <typename T>
 void declare_meshtags(nb::module_& m, std::string type)
 {
@@ -122,8 +94,8 @@ void declare_meshtags(nb::module_& m, std::string type)
           "values",
           [](dolfinx::mesh::MeshTags<T>& self)
           {
-            return nb::ndarray<const T, nb::numpy>(self.values().data(),
-                                                   {self.values().size()});
+            return nb::ndarray<const T, nb::numpy>(
+                self.values().data(), {self.values().size()}, nb::handle());
           },
           nb::rv_policy::reference_internal)
       .def_prop_ro(
@@ -131,7 +103,7 @@ void declare_meshtags(nb::module_& m, std::string type)
           [](dolfinx::mesh::MeshTags<T>& self)
           {
             return nb::ndarray<const std::int32_t, nb::numpy>(
-                self.indices().data(), {self.indices().size()});
+                self.indices().data(), {self.indices().size()}, nb::handle());
           },
           nb::rv_policy::reference_internal)
       .def("find", [](dolfinx::mesh::MeshTags<T>& self, T value)
@@ -153,6 +125,40 @@ void declare_mesh(nb::module_& m, std::string type)
   std::string pyclass_geometry_name = std::string("Geometry_") + type;
   nb::class_<dolfinx::mesh::Geometry<T>>(m, pyclass_geometry_name.c_str(),
                                          "Geometry object")
+      .def(
+          "__init__",
+          [](dolfinx::mesh::Geometry<T>* self,
+             std::shared_ptr<const dolfinx::common::IndexMap> index_map,
+             nb::ndarray<const std::int32_t, nb::ndim<2>, nb::c_contig> dofmap,
+             const dolfinx::fem::CoordinateElement<T>& element,
+             nb::ndarray<const T, nb::ndim<2>> x,
+             nb::ndarray<const std::int64_t, nb::ndim<1>, nb::c_contig>
+                 input_global_indices)
+          {
+            int shape1 = x.shape(1);
+            std::vector<T> x_vec;
+            if (shape1 == 3 and x.stride(0) == 3 and x.stride(1) == 1)
+              x_vec.assign(x.data(), x.data() + x.size());
+            else
+            {
+              // Pad geometry to be 3D
+              x_vec.assign(3 * x.shape(0), 0);
+              auto _x = x.view();
+              for (std::size_t i = 0; i < x.shape(0); ++i)
+                for (int j = 0; j < shape1; ++j)
+                  x_vec[3 * i + j] = _x(i, j);
+            }
+
+            new (self) dolfinx::mesh::Geometry<T>(
+                index_map,
+                std::vector(dofmap.data(), dofmap.data() + dofmap.size()),
+                element, std::move(x_vec), shape1,
+                std::vector(input_global_indices.data(),
+                            input_global_indices.data()
+                                + input_global_indices.size()));
+          },
+          nb::arg("index_map"), nb::arg("dofmap"), nb::arg("element"),
+          nb::arg("x"), nb::arg("input_global_indices"))
       .def_prop_ro("dim", &dolfinx::mesh::Geometry<T>::dim,
                    "Geometric dimension")
       .def_prop_ro(
@@ -161,7 +167,8 @@ void declare_mesh(nb::module_& m, std::string type)
           {
             auto dofs = self.dofmap();
             return nb::ndarray<const std::int32_t, nb::numpy>(
-                dofs.data_handle(), {dofs.extent(0), dofs.extent(1)});
+                dofs.data_handle(), {dofs.extent(0), dofs.extent(1)},
+                nb::handle());
           },
           nb::rv_policy::reference_internal)
       .def(
@@ -170,7 +177,8 @@ void declare_mesh(nb::module_& m, std::string type)
           {
             auto dofs = self.dofmap(i);
             return nb::ndarray<const std::int32_t, nb::numpy>(
-                dofs.data_handle(), {dofs.extent(0), dofs.extent(1)});
+                dofs.data_handle(), {dofs.extent(0), dofs.extent(1)},
+                nb::handle());
           },
           nb::rv_policy::reference_internal, nb::arg("i"),
           "Get the geometry dofmap associated with coordinate element i (mixed "
@@ -180,8 +188,8 @@ void declare_mesh(nb::module_& m, std::string type)
           "x",
           [](dolfinx::mesh::Geometry<T>& self)
           {
-            return nb::ndarray<T, nb::numpy>(self.x().data(),
-                                             {self.x().size() / 3, 3});
+            return nb::ndarray<T, nb::numpy>(
+                self.x().data(), {self.x().size() / 3, 3}, nb::handle());
           },
           nb::rv_policy::reference_internal,
           "Return coordinates of all geometry points. Each row is the "
@@ -189,6 +197,9 @@ void declare_mesh(nb::module_& m, std::string type)
       .def_prop_ro(
           "cmap", [](dolfinx::mesh::Geometry<T>& self) { return self.cmap(); },
           "The coordinate map")
+      .def(
+          "cmaps", [](dolfinx::mesh::Geometry<T>& self, int i)
+          { return self.cmap(i); }, "The ith coordinate map")
       .def_prop_ro(
           "input_global_indices",
           [](const dolfinx::mesh::Geometry<T>& self)
@@ -196,7 +207,7 @@ void declare_mesh(nb::module_& m, std::string type)
             const std::vector<std::int64_t>& id_to_global
                 = self.input_global_indices();
             return nb::ndarray<const std::int64_t, nb::numpy>(
-                id_to_global.data(), {id_to_global.size()});
+                id_to_global.data(), {id_to_global.size()}, nb::handle());
           },
           nb::rv_policy::reference_internal);
 
@@ -207,7 +218,8 @@ void declare_mesh(nb::module_& m, std::string type)
           "__init__",
           [](dolfinx::mesh::Mesh<T>* mesh, MPICommWrapper comm,
              std::shared_ptr<dolfinx::mesh::Topology> topology,
-             dolfinx::mesh::Geometry<T>& geometry) {
+             dolfinx::mesh::Geometry<T>& geometry)
+          {
             new (mesh) dolfinx::mesh::Mesh<T>(comm.get(), topology, geometry);
           },
           nb::arg("comm"), nb::arg("topology"), nb::arg("geometry"))
@@ -225,12 +237,13 @@ void declare_mesh(nb::module_& m, std::string type)
   std::string create_interval("create_interval_" + type);
   m.def(
       create_interval.c_str(),
-      [](MPICommWrapper comm, std::size_t n, std::array<double, 2> p,
-         dolfinx::mesh::GhostMode ghost_mode,
-         const PythonCellPartitionFunction& part)
+      [](MPICommWrapper comm, std::int64_t n, std::array<T, 2> p,
+         dolfinx::mesh::GhostMode mode,
+         const part::impl::PythonCellPartitionFunction& part)
       {
         return dolfinx::mesh::create_interval<T>(
-            comm.get(), n, p, create_cell_partitioner_cpp(part));
+            comm.get(), n, p, mode,
+            part::impl::create_cell_partitioner_cpp(part));
       },
       nb::arg("comm"), nb::arg("n"), nb::arg("p"), nb::arg("ghost_mode"),
       nb::arg("partitioner").none());
@@ -238,14 +251,14 @@ void declare_mesh(nb::module_& m, std::string type)
   std::string create_rectangle("create_rectangle_" + type);
   m.def(
       create_rectangle.c_str(),
-      [](MPICommWrapper comm, std::array<std::array<double, 2>, 2> p,
-         std::array<std::size_t, 2> n, dolfinx::mesh::CellType celltype,
-         const PythonCellPartitionFunction& part,
+      [](MPICommWrapper comm, std::array<std::array<T, 2>, 2> p,
+         std::array<std::int64_t, 2> n, dolfinx::mesh::CellType celltype,
+         const part::impl::PythonCellPartitionFunction& part,
          dolfinx::mesh::DiagonalType diagonal)
       {
         return dolfinx::mesh::create_rectangle<T>(
-            comm.get(), p, n, celltype, create_cell_partitioner_cpp(part),
-            diagonal);
+            comm.get(), p, n, celltype,
+            part::impl::create_cell_partitioner_cpp(part), diagonal);
       },
       nb::arg("comm"), nb::arg("p"), nb::arg("n"), nb::arg("celltype"),
       nb::arg("partitioner").none(), nb::arg("diagonal"));
@@ -253,31 +266,86 @@ void declare_mesh(nb::module_& m, std::string type)
   std::string create_box("create_box_" + type);
   m.def(
       create_box.c_str(),
-      [](MPICommWrapper comm, std::array<std::array<double, 3>, 2> p,
-         std::array<std::size_t, 3> n, dolfinx::mesh::CellType celltype,
-         const PythonCellPartitionFunction& part)
+      [](MPICommWrapper comm, std::array<std::array<T, 3>, 2> p,
+         std::array<std::int64_t, 3> n, dolfinx::mesh::CellType celltype,
+         const part::impl::PythonCellPartitionFunction& part)
       {
         MPI_Comm _comm = comm.get();
-        return dolfinx::mesh::create_box<T>(_comm, _comm, p, n, celltype,
-                                            create_cell_partitioner_cpp(part));
+        return dolfinx::mesh::create_box<T>(
+            _comm, _comm, p, n, celltype,
+            part::impl::create_cell_partitioner_cpp(part));
       },
       nb::arg("comm"), nb::arg("p"), nb::arg("n"), nb::arg("celltype"),
       nb::arg("partitioner").none());
+
+  m.def("create_mesh",
+        [](MPICommWrapper comm,
+           const std::vector<nb::ndarray<const std::int64_t, nb::ndim<1>,
+                                         nb::c_contig>>& cells_nb,
+           const std::vector<dolfinx::fem::CoordinateElement<T>>& elements,
+           nb::ndarray<const T, nb::c_contig> x,
+           const part::impl::PythonCellPartitionFunction& p)
+        {
+          std::size_t shape1 = x.ndim() == 1 ? 1 : x.shape(1);
+
+          std::vector<std::span<const std::int64_t>> cells;
+          std::ranges::transform(
+              cells_nb, std::back_inserter(cells), [](auto c)
+              { return std::span<const std::int64_t>(c.data(), c.size()); });
+
+          if (p)
+          {
+            auto p_wrap
+                = [p](MPI_Comm comm, int n,
+                      const std::vector<dolfinx::mesh::CellType>& cell_types,
+                      const std::vector<std::span<const std::int64_t>>& cells)
+            {
+              std::vector<nb::ndarray<const std::int64_t, nb::numpy>> cells_nb;
+              std::ranges::transform(
+                  cells, std::back_inserter(cells_nb),
+                  [](auto c)
+                  {
+                    return nb::ndarray<const std::int64_t, nb::numpy>(
+                        c.data(), {c.size()}, nb::handle());
+                  });
+              return p(MPICommWrapper(comm), n, cell_types, cells_nb);
+            };
+            return dolfinx::mesh::create_mesh(
+                comm.get(), comm.get(), cells, elements, comm.get(),
+                std::span(x.data(), x.size()), {x.shape(0), shape1}, p_wrap);
+          }
+          else
+            return dolfinx::mesh::create_mesh(
+                comm.get(), comm.get(), cells, elements, comm.get(),
+                std::span(x.data(), x.size()), {x.shape(0), shape1}, nullptr);
+        });
+
   m.def(
       "create_mesh",
       [](MPICommWrapper comm,
          nb::ndarray<const std::int64_t, nb::ndim<2>, nb::c_contig> cells,
          const dolfinx::fem::CoordinateElement<T>& element,
          nb::ndarray<const T, nb::c_contig> x,
-         const PythonPartitioningFunction& p)
+         const part::impl::PythonCellPartitionFunction& p)
       {
         std::size_t shape1 = x.ndim() == 1 ? 1 : x.shape(1);
         if (p)
         {
           auto p_wrap
-              = [p](MPI_Comm comm, int n, dolfinx::mesh::CellType cell_type,
-                    const dolfinx::graph::AdjacencyList<std::int64_t>& cells)
-          { return p(MPICommWrapper(comm), n, cell_type, cells); };
+              = [p](MPI_Comm comm, int n,
+                    const std::vector<dolfinx::mesh::CellType>& cell_types,
+                    const std::vector<std::span<const std::int64_t>>& cells)
+          {
+            std::vector<nb::ndarray<const std::int64_t, nb::numpy>> cells_nb;
+            std::ranges::transform(
+                cells, std::back_inserter(cells_nb),
+                [](auto c)
+                {
+                  return nb::ndarray<const std::int64_t, nb::numpy>(
+                      c.data(), {c.size()}, nb::handle());
+                });
+            return p(MPICommWrapper(comm), n, cell_types, cells_nb);
+          };
           return dolfinx::mesh::create_mesh(
               comm.get(), comm.get(), std::span(cells.data(), cells.size()),
               element, comm.get(), std::span(x.data(), x.size()),
@@ -288,7 +356,7 @@ void declare_mesh(nb::module_& m, std::string type)
           return dolfinx::mesh::create_mesh(
               comm.get(), comm.get(), std::span(cells.data(), cells.size()),
               element, comm.get(), std::span(x.data(), x.size()),
-              {x.shape(0), shape1}, p);
+              {x.shape(0), shape1}, nullptr);
         }
       },
       nb::arg("comm"), nb::arg("cells"), nb::arg("element"),
@@ -350,7 +418,7 @@ void declare_mesh(nb::module_& m, std::string type)
         auto cpp_marker = [&marker](auto x)
         {
           nb::ndarray<const T, nb::ndim<2>, nb::numpy> x_view(
-              x.data_handle(), {x.extent(0), x.extent(1)});
+              x.data_handle(), {x.extent(0), x.extent(1)}, nb::handle());
           auto marked = marker(x_view);
           return std::vector<std::int8_t>(marked.data(),
                                           marked.data() + marked.size());
@@ -371,7 +439,7 @@ void declare_mesh(nb::module_& m, std::string type)
         auto cpp_marker = [&marker](auto x)
         {
           nb::ndarray<const T, nb::ndim<2>, nb::numpy> x_view(
-              x.data_handle(), {x.extent(0), x.extent(1)});
+              x.data_handle(), {x.extent(0), x.extent(1)}, nb::handle());
           auto marked = marker(x_view);
           return std::vector<std::int8_t>(marked.data(),
                                           marked.data() + marked.size());
@@ -385,19 +453,18 @@ void declare_mesh(nb::module_& m, std::string type)
       "entities_to_geometry",
       [](const dolfinx::mesh::Mesh<T>& mesh, int dim,
          nb::ndarray<const std::int32_t, nb::ndim<1>, nb::c_contig> entities,
-         bool orient)
+         bool permute)
       {
         std::vector<std::int32_t> idx = dolfinx::mesh::entities_to_geometry(
-            mesh, dim, std::span(entities.data(), entities.size()), orient);
+            mesh, dim, std::span(entities.data(), entities.size()), permute);
 
         auto topology = mesh.topology();
         assert(topology);
         dolfinx::mesh::CellType cell_type = topology->cell_type();
-        std::size_t num_vertices = dolfinx::mesh::num_cell_vertices(
-            cell_entity_type(cell_type, dim, 0));
-        return as_nbarray(std::move(idx), {entities.size(), num_vertices});
+        return as_nbarray(std::move(idx),
+                          {entities.size(), idx.size() / entities.size()});
       },
-      nb::arg("mesh"), nb::arg("dim"), nb::arg("entities"), nb::arg("orient"));
+      nb::arg("mesh"), nb::arg("dim"), nb::arg("entities"), nb::arg("permute"));
 
   m.def("create_geometry",
         [](const dolfinx::mesh::Topology& topology,
@@ -456,10 +523,33 @@ void mesh(nb::module_& m)
       "build_dual_graph",
       [](const MPICommWrapper comm, dolfinx::mesh::CellType cell_type,
          const dolfinx::graph::AdjacencyList<std::int64_t>& cells)
-      { return dolfinx::mesh::build_dual_graph(comm.get(), cell_type, cells); },
+      {
+        std::vector<dolfinx::mesh::CellType> c = {cell_type};
+        return dolfinx::mesh::build_dual_graph(comm.get(), std::span{c},
+                                               {cells.array()});
+      },
       nb::arg("comm"), nb::arg("cell_type"), nb::arg("cells"),
       "Build dual graph for cells");
 
+  m.def(
+      "build_dual_graph",
+      [](const MPICommWrapper comm,
+         std::vector<dolfinx::mesh::CellType>& cell_types,
+         const std::vector<
+             nb::ndarray<const std::int64_t, nb::ndim<1>, nb::c_contig>>& cells)
+      {
+        std::vector<std::span<const std::int64_t>> cell_span(cells.size());
+        for (std::size_t i = 0; i < cells.size(); ++i)
+        {
+          cell_span[i]
+              = std::span<const std::int64_t>(cells[i].data(), cells[i].size());
+        }
+        return dolfinx::mesh::build_dual_graph(comm.get(), cell_types,
+                                               cell_span);
+      },
+      nb::arg("comm"), nb::arg("cell_types"), nb::arg("cells"),
+      "Build dual graph for cells");
+
   // dolfinx::mesh::GhostMode enums
   nb::enum_<dolfinx::mesh::GhostMode>(m, "GhostMode")
       .value("none", dolfinx::mesh::GhostMode::none)
@@ -470,7 +560,8 @@ void mesh(nb::module_& m)
   m.def(
       "compute_entities",
       [](MPICommWrapper comm, const dolfinx::mesh::Topology& topology, int dim,
-         int index) {
+         int index)
+      {
         return dolfinx::mesh::compute_entities(comm.get(), topology, dim,
                                                index);
       },
@@ -508,8 +599,8 @@ void mesh(nb::module_& m)
           [](const dolfinx::mesh::Topology& self)
           {
             const std::vector<std::uint8_t>& p = self.get_facet_permutations();
-            return nb::ndarray<const std::uint8_t, nb::numpy>(p.data(),
-                                                              {p.size()});
+            return nb::ndarray<const std::uint8_t, nb::numpy>(
+                p.data(), {p.size()}, nb::handle());
           },
           nb::rv_policy::reference_internal)
       .def(
@@ -518,8 +609,8 @@ void mesh(nb::module_& m)
           {
             const std::vector<std::uint32_t>& p
                 = self.get_cell_permutation_info();
-            return nb::ndarray<const std::uint32_t, nb::numpy>(p.data(),
-                                                               {p.size()});
+            return nb::ndarray<const std::uint32_t, nb::numpy>(
+                p.data(), {p.size()}, nb::handle());
           },
           nb::rv_policy::reference_internal)
       .def_prop_ro("dim", &dolfinx::mesh::Topology::dim,
@@ -532,7 +623,7 @@ void mesh(nb::module_& m)
               throw std::runtime_error("Mixed topology unsupported");
             return nb::ndarray<const std::int64_t, nb::numpy>(
                 self.original_cell_index[0].data(),
-                {self.original_cell_index[0].size()});
+                {self.original_cell_index[0].size()}, nb::handle());
           },
           nb::rv_policy::reference_internal)
       .def("connectivity",
@@ -556,14 +647,16 @@ void mesh(nb::module_& m)
               entity_types.push_back(self.entity_types(i));
             return entity_types;
           })
-      .def("cell_name", [](const dolfinx::mesh::Topology& self)
-           { return dolfinx::mesh::to_string(self.cell_type()); })
-      .def("interprocess_facets",
-           nb::overload_cast<>(&dolfinx::mesh::Topology::interprocess_facets,
-                               nb::const_))
-      .def("interprocess_facets",
-           nb::overload_cast<std::int8_t>(
-               &dolfinx::mesh::Topology::interprocess_facets, nb::const_))
+      .def(
+          "interprocess_facets",
+          [](const dolfinx::mesh::Topology& self)
+          {
+            const std::vector<std::int32_t>& facets
+                = self.interprocess_facets();
+            return nb::ndarray<const std::int32_t, nb::numpy>(
+                facets.data(), {facets.size()}, nb::handle());
+          },
+          nb::rv_policy::reference_internal)
       .def_prop_ro(
           "comm", [](dolfinx::mesh::Topology& self)
           { return MPICommWrapper(self.comm()); }, nb::keep_alive<0, 1>());
@@ -590,8 +683,6 @@ void mesh(nb::module_& m)
               ghost_owners_span, boundary_vertices_span);
         });
 
-  // dolfinx::mesh::MeshTags
-
   declare_meshtags<std::int8_t>(m, "int8");
   declare_meshtags<std::int32_t>(m, "int32");
   declare_meshtags<std::int64_t>(m, "int64");
@@ -602,10 +693,11 @@ void mesh(nb::module_& m)
 
   m.def(
       "create_cell_partitioner",
-      [](dolfinx::mesh::GhostMode gm) -> PythonCellPartitionFunction
+      [](dolfinx::mesh::GhostMode mode)
+          -> part::impl::PythonCellPartitionFunction
       {
-        return create_cell_partitioner_py(
-            dolfinx::mesh::create_cell_partitioner(gm));
+        return part::impl::create_cell_partitioner_py(
+            dolfinx::mesh::create_cell_partitioner(mode));
       },
       "Create default cell partitioner.");
   m.def(
@@ -615,11 +707,12 @@ void mesh(nb::module_& m)
              const dolfinx::graph::AdjacencyList<std::int64_t>& local_graph,
              bool ghosting)>
              part,
-         dolfinx::mesh::GhostMode ghost_mode) -> PythonCellPartitionFunction
+         dolfinx::mesh::GhostMode mode)
+          -> part::impl::PythonCellPartitionFunction
       {
-        return create_cell_partitioner_py(
+        return part::impl::create_cell_partitioner_py(
             dolfinx::mesh::create_cell_partitioner(
-                ghost_mode, create_partitioner_cpp(part)));
+                mode, part::impl::create_partitioner_cpp(part)));
       },
       nb::arg("part"), nb::arg("ghost_mode") = dolfinx::mesh::GhostMode::none,
       "Create a cell partitioner from a graph partitioning function.");
