@@ -19,13 +19,14 @@ py_to_cpp_coeffs(
 {
   using Key_t = typename std::remove_reference_t<decltype(coeffs)>::key_type;
   std::map<Key_t, std::pair<std::span<const T>, int>> c;
-  std::transform(coeffs.begin(), coeffs.end(), std::inserter(c, c.end()),
-                 [](auto& e) -> typename decltype(c)::value_type
-                 {
-                   return {e.first,
-                           {std::span(static_cast<const T*>(e.second.data()),
-                                      e.second.shape(0)),
-                            e.second.shape(1)}};
-                 });
+  std::ranges::transform(
+      coeffs, std::inserter(c, c.end()),
+      [](auto& e) -> typename decltype(c)::value_type
+      {
+        return {e.first,
+                {std::span(static_cast<const T*>(e.second.data()),
+                           e.second.shape(0)),
+                 e.second.shape(1)}};
+      });
   return c;
 }
