@@ -38,6 +38,16 @@ parametrize_cell_types = pytest.mark.parametrize(
     ],
 )
 
+parametrize_dtypes = pytest.mark.parametrize(
+    "dtype",
+    [
+        np.float32,
+        np.float64,
+        pytest.param(np.complex64, marks=pytest.mark.xfail_win32_complex),
+        pytest.param(np.complex128, marks=pytest.mark.xfail_win32_complex),
+    ],
+)
+
 
 def unit_cell_points(cell_type, dtype):
     if cell_type == CellType.interval:
@@ -186,7 +196,7 @@ def two_unit_cells(cell_type, dtype, agree=False, random_order=True, return_orde
 
 @pytest.mark.skip_in_parallel
 @parametrize_cell_types
-@pytest.mark.parametrize("dtype", [np.float32, np.float64, np.complex64, np.complex128])
+@parametrize_dtypes
 def test_facet_integral(cell_type, dtype):
     """Test that the integral of a function over a facet is correct"""
     xtype = np.real(dtype(0)).dtype
@@ -229,12 +239,12 @@ def test_facet_integral(cell_type, dtype):
             a = form(v * ufl.ds(subdomain_data=marker, subdomain_id=j), dtype=dtype)
             result = assemble_scalar(a)
             out.append(result)
-            assert np.isclose(result, out[0])
+            assert np.isclose(result, out[0], atol=np.finfo(dtype).eps)
 
 
 @pytest.mark.skip_in_parallel
 @parametrize_cell_types
-@pytest.mark.parametrize("dtype", [np.float32, np.float64, np.complex64, np.complex128])
+@parametrize_dtypes
 def test_facet_normals(cell_type, dtype):
     """Test that FacetNormal is outward facing"""
     xtype = np.real(dtype(0)).dtype
@@ -300,7 +310,7 @@ def test_facet_normals(cell_type, dtype):
                     dtype=dtype,
                 )
                 result = assemble_scalar(a)
-                if np.isclose(result, 1):
+                if np.isclose(result, 1, atol=np.finfo(dtype).eps):
                     ones += 1
                 else:
                     assert np.isclose(result, 0, atol=1.0e-6)
@@ -310,7 +320,7 @@ def test_facet_normals(cell_type, dtype):
 @pytest.mark.skip_in_parallel
 @pytest.mark.parametrize("space_type", ["Lagrange", "DG"])
 @parametrize_cell_types
-@pytest.mark.parametrize("dtype", [np.float32, np.float64, np.complex64, np.complex128])
+@parametrize_dtypes
 def test_plus_minus(cell_type, space_type, dtype):
     """Test that ('+') and ('-') give the same value for continuous functions"""
     xtype = np.real(dtype(0)).dtype
@@ -326,13 +336,13 @@ def test_plus_minus(cell_type, space_type, dtype):
                 a = form(v(pm1) * v(pm2) * ufl.dS, dtype=dtype)
                 results.append(assemble_scalar(a))
     for i, j in combinations(results, 2):
-        assert np.isclose(i, j)
+        assert np.isclose(i, j, atol=np.finfo(dtype).eps)
 
 
 @pytest.mark.skip_in_parallel
 @pytest.mark.parametrize("pm", ["+", "-"])
 @parametrize_cell_types
-@pytest.mark.parametrize("dtype", [np.float32, np.float64, np.complex64, np.complex128])
+@parametrize_dtypes
 def test_plus_minus_simple_vector(cell_type, pm, dtype):
     """Test that ('+') and ('-') match up with the correct DOFs for DG functions"""
     xtype = np.real(dtype(0)).dtype
@@ -378,14 +388,14 @@ def test_plus_minus_simple_vector(cell_type, pm, dtype):
                     # If no matching point found, fail
                     assert False
 
-                assert np.isclose(results[0][dof0], result[dof1])
+                assert np.isclose(results[0][dof0], result[dof1], atol=np.finfo(dtype).eps)
 
 
 @pytest.mark.skip_in_parallel
 @pytest.mark.parametrize("pm1", ["+", "-"])
 @pytest.mark.parametrize("pm2", ["+", "-"])
 @parametrize_cell_types
-@pytest.mark.parametrize("dtype", [np.float32, np.float64, np.complex64, np.complex128])
+@parametrize_dtypes
 def test_plus_minus_vector(cell_type, pm1, pm2, dtype):
     """Test that ('+') and ('-') match up with the correct DOFs for DG functions"""
     xtype = np.real(dtype(0)).dtype
@@ -440,7 +450,7 @@ def test_plus_minus_vector(cell_type, pm1, pm2, dtype):
 @pytest.mark.parametrize("pm1", ["+", "-"])
 @pytest.mark.parametrize("pm2", ["+", "-"])
 @parametrize_cell_types
-@pytest.mark.parametrize("dtype", [np.float32, np.float64, np.complex64, np.complex128])
+@parametrize_dtypes
 def test_plus_minus_matrix(cell_type, pm1, pm2, dtype):
     """Test that ('+') and ('-') match up with the correct DOFs for DG functions"""
     xtype = np.real(dtype(0)).dtype
@@ -490,7 +500,7 @@ def test_plus_minus_matrix(cell_type, pm1, pm2, dtype):
         # For all dof pairs, check that entries in the matrix agree
         for a, b in dof_order:
             for c, d in dof_order:
-                assert np.isclose(results[0][a, c], result[b, d])
+                assert np.isclose(results[0][a, c], result[b, d], atol=np.finfo(dtype).eps)
 
 
 @pytest.mark.skip(
@@ -499,7 +509,7 @@ def test_plus_minus_matrix(cell_type, pm1, pm2, dtype):
 @pytest.mark.skip_in_parallel
 @pytest.mark.parametrize("order", [1, 2])
 @pytest.mark.parametrize("space_type", ["N1curl", "N2curl"])
-@pytest.mark.parametrize("dtype", [np.float32, np.float64, np.complex64, np.complex128])
+@parametrize_dtypes
 def test_curl(space_type, order, dtype):
     """Test that curl is consistent for different cell permutations of a tetrahedron."""
     xtype = np.real(dtype(0)).dtype
@@ -572,7 +582,7 @@ def create_quad_mesh(offset, dtype):
 
 @pytest.mark.skip_in_parallel
 @pytest.mark.parametrize("k", [0, 1, 2])
-@pytest.mark.parametrize("dtype", [np.float32, np.float64, np.complex64, np.complex128])
+@parametrize_dtypes
 def test_div_general_quads_mat(k, dtype):
     """Tests that assembling inner(u, div(w)) * dx, where u is from a
     "DQ" space and w is from an "RTCF" space, gives the same matrix for
@@ -601,7 +611,7 @@ def test_div_general_quads_mat(k, dtype):
 
 @pytest.mark.skip_in_parallel
 @pytest.mark.parametrize("k", [0, 1, 2])
-@pytest.mark.parametrize("dtype", [np.float32, np.float64, np.complex64, np.complex128])
+@parametrize_dtypes
 def test_div_general_quads_vec(k, dtype):
     """Tests that assembling inner(1, div(w)) * dx, where w is from an
     "RTCF" space, gives the same matrix for square and trapezoidal
