@@ -20,7 +20,7 @@ from dolfinx.fem import (
     Expression,
     Function,
     assemble_scalar,
-    create_nonmatching_meshes_interpolation_data,
+    create_interpolation_data,
     form,
     functionspace,
 )
@@ -345,8 +345,10 @@ def test_mixed_sub_interpolation():
     def f(x):
         return np.vstack((10 + x[0], -10 - x[1], 25 + x[0]))
 
-    P2 = element("Lagrange", mesh.basix_cell(), 2, shape=(mesh.geometry.dim,))
-    P1 = element("Lagrange", mesh.basix_cell(), 1)
+    P2 = element(
+        "Lagrange", mesh.basix_cell(), 2, shape=(mesh.geometry.dim,), dtype=default_real_type
+    )
+    P1 = element("Lagrange", mesh.basix_cell(), 1, dtype=default_real_type)
     for i, P in enumerate((mixed_element([P2, P1]), mixed_element([P1, P2]))):
         W = functionspace(mesh, P)
         U = Function(W)
@@ -397,8 +399,10 @@ def test_mixed_sub_interpolation():
 def test_mixed_interpolation():
     """Test that mixed interpolation raised an exception."""
     mesh = one_cell_mesh(CellType.triangle)
-    A = element("Lagrange", mesh.basix_cell(), 1)
-    B = element("Lagrange", mesh.basix_cell(), 1, shape=(mesh.geometry.dim,))
+    A = element("Lagrange", mesh.basix_cell(), 1, dtype=default_real_type)
+    B = element(
+        "Lagrange", mesh.basix_cell(), 1, shape=(mesh.geometry.dim,), dtype=default_real_type
+    )
     v = Function(functionspace(mesh, mixed_element([A, B])))
     with pytest.raises(RuntimeError):
         v.interpolate(lambda x: (x[1], 2 * x[0], 3 * x[1]))
@@ -679,7 +683,7 @@ def test_de_rahm_2D(order):
     x = ufl.SpatialCoordinate(mesh)
     g_ex = ufl.as_vector((1 + x[1], 4 * x[1] + x[0]))
     assert np.abs(assemble_scalar(form(ufl.inner(q - g_ex, q - g_ex) * ufl.dx))) == pytest.approx(
-        0, abs=1e-10
+        0, abs=np.sqrt(np.finfo(mesh.geometry.x.dtype).eps)
     )
 
     V = functionspace(mesh, ("BDM", order - 1))
@@ -691,7 +695,7 @@ def test_de_rahm_2D(order):
     v.interpolate(Expression(curl2D(ufl.grad(w)), V.element.interpolation_points()))
     h_ex = ufl.as_vector((1, -1))
     assert np.abs(assemble_scalar(form(ufl.inner(v - h_ex, v - h_ex) * ufl.dx))) == pytest.approx(
-        0, abs=1.0e-6
+        0, abs=np.sqrt(np.finfo(mesh.geometry.x.dtype).eps)
     )
 
 
@@ -765,17 +769,27 @@ def test_interpolate_callable_subset(bound):
 @pytest.mark.parametrize(
     "scalar_element",
     [
-        element("P", "triangle", 1),
-        element("P", "triangle", 2),
-        element("P", "triangle", 3),
-        element("Q", "quadrilateral", 1),
-        element("Q", "quadrilateral", 2),
-        element("Q", "quadrilateral", 3),
-        element("S", "quadrilateral", 1),
-        element("S", "quadrilateral", 2),
-        element("S", "quadrilateral", 3),
-        enriched_element([element("P", "triangle", 1), element("Bubble", "triangle", 3)]),
-        enriched_element([element("P", "quadrilateral", 1), element("Bubble", "quadrilateral", 2)]),
+        element("P", "triangle", 1, dtype=default_real_type),
+        element("P", "triangle", 2, dtype=default_real_type),
+        element("P", "triangle", 3, dtype=default_real_type),
+        element("Q", "quadrilateral", 1, dtype=default_real_type),
+        element("Q", "quadrilateral", 2, dtype=default_real_type),
+        element("Q", "quadrilateral", 3, dtype=default_real_type),
+        element("S", "quadrilateral", 1, dtype=default_real_type),
+        element("S", "quadrilateral", 2, dtype=default_real_type),
+        element("S", "quadrilateral", 3, dtype=default_real_type),
+        enriched_element(
+            [
+                element("P", "triangle", 1, dtype=default_real_type),
+                element("Bubble", "triangle", 3, dtype=default_real_type),
+            ]
+        ),
+        enriched_element(
+            [
+                element("P", "quadrilateral", 1, dtype=default_real_type),
+                element("Bubble", "quadrilateral", 2, dtype=default_real_type),
+            ]
+        ),
     ],
 )
 def test_vector_element_interpolation(scalar_element):
@@ -821,6 +835,7 @@ def test_custom_vector_element():
         False,
         1,
         1,
+        dtype=default_real_type,
     )
 
     V = functionspace(mesh, e)
@@ -846,9 +861,11 @@ def test_mixed_interpolation_permuting(cell_type, order):
     x = ufl.SpatialCoordinate(mesh)
     dgdy = ufl.cos(x[1])
 
-    curl_el = element("N1curl", mesh.basix_cell(), 1)
-    vlag_el = element("Lagrange", mesh.basix_cell(), 1, shape=(mesh.geometry.dim,))
-    lagr_el = element("Lagrange", mesh.basix_cell(), order)
+    curl_el = element("N1curl", mesh.basix_cell(), 1, dtype=default_real_type)
+    vlag_el = element(
+        "Lagrange", mesh.basix_cell(), 1, shape=(mesh.geometry.dim,), dtype=default_real_type
+    )
+    lagr_el = element("Lagrange", mesh.basix_cell(), order, dtype=default_real_type)
 
     V = functionspace(mesh, mixed_element([curl_el, lagr_el]))
     Eb_m = Function(V)
@@ -873,9 +890,9 @@ def test_nonmatching_mesh_interpolation(xtype, cell_type0, cell_type1):
     def f(x):
         return (7 * x[1], 3 * x[0], x[2] + 0.4)
 
-    el0 = element("Lagrange", mesh0.basix_cell(), 1, shape=(3,))
+    el0 = element("Lagrange", mesh0.basix_cell(), 1, shape=(3,), dtype=xtype)
     V0 = functionspace(mesh0, el0)
-    el1 = element("Lagrange", mesh1.basix_cell(), 1, shape=(3,))
+    el1 = element("Lagrange", mesh1.basix_cell(), 1, shape=(3,), dtype=xtype)
     V1 = functionspace(mesh1, el1)
 
     # Interpolate on 3D mesh
@@ -888,22 +905,12 @@ def test_nonmatching_mesh_interpolation(xtype, cell_type0, cell_type1):
     fine_mesh_cell_map = mesh1.topology.index_map(mesh1.topology.dim)
     num_cells_on_proc = fine_mesh_cell_map.size_local + fine_mesh_cell_map.num_ghosts
     cells = np.arange(num_cells_on_proc, dtype=np.int32)
-    interpolation_data = create_nonmatching_meshes_interpolation_data(
-        V1.mesh.geometry, V1.element, V0.mesh, cells, padding=padding
-    )
-    other_interpolation_data = create_nonmatching_meshes_interpolation_data(
-        V1.mesh,
-        V1.element,
-        V0.mesh,
-        padding=padding,
-    )
-    for data_0, data_1 in zip(interpolation_data, other_interpolation_data):
-        np.testing.assert_allclose(data_0, data_1)
+    interpolation_data = create_interpolation_data(V1, V0, cells, padding=padding)
 
     # Interpolate 3D->2D
     u1 = Function(V1, dtype=xtype)
 
-    u1.interpolate(u0, nmm_interpolation_data=interpolation_data)
+    u1.interpolate_nonmatching(u0, cells, interpolation_data=interpolation_data)
     u1.x.scatter_forward()
 
     # Exact interpolation on 2D mesh
@@ -911,19 +918,20 @@ def test_nonmatching_mesh_interpolation(xtype, cell_type0, cell_type1):
     u1_ex.interpolate(f)
     u1_ex.x.scatter_forward()
 
-    assert np.allclose(u1_ex.x.array, u1.x.array, rtol=1.0e-4, atol=1.0e-6)
+    assert np.allclose(
+        u1_ex.x.array,
+        u1.x.array,
+        rtol=np.sqrt(np.finfo(xtype).eps),
+        atol=np.sqrt(np.finfo(xtype).eps),
+    )
 
     # Interpolate 2D->3D
+    cell_map0 = mesh0.topology.index_map(mesh0.topology.dim)
+    num_cells_on_proc = cell_map0.size_local + cell_map0.num_ghosts
+    cells0 = np.arange(num_cells_on_proc, dtype=np.int32)
+    interpolation_data1 = create_interpolation_data(V0, V1, cells0, padding=padding)
     u0_2 = Function(V0, dtype=xtype)
-    u0_2.interpolate(
-        u1,
-        nmm_interpolation_data=create_nonmatching_meshes_interpolation_data(
-            u0_2.function_space.mesh,
-            u0_2.function_space.element,
-            u1.function_space.mesh,
-            padding=padding,
-        ),
-    )
+    u0_2.interpolate_nonmatching(u1, cells0, interpolation_data1)
 
     # Check that function values over facets of 3D mesh of the twice
     # interpolated property is preserved
@@ -974,11 +982,14 @@ def test_nonmatching_mesh_single_cell_overlap_interpolation(xtype):
     u1.interpolate(f_test1)
     u1.x.scatter_forward()
     padding = 1e-14
-    u1_2_u2_nmm_data = create_nonmatching_meshes_interpolation_data(
-        u2.function_space.mesh, u2.function_space.element, u1.function_space.mesh, padding=padding
+    cell_map2 = mesh2.topology.index_map(mesh2.topology.dim)
+    num_cells2 = cell_map2.size_local + cell_map2.num_ghosts
+    cells2 = np.arange(num_cells2, dtype=np.int32)
+    u1_2_u2_nmm_data = create_interpolation_data(
+        u2.function_space, u1.function_space, cells2, padding=padding
     )
 
-    u2.interpolate(u1, nmm_interpolation_data=u1_2_u2_nmm_data)
+    u2.interpolate_nonmatching(u1, cells2, interpolation_data=u1_2_u2_nmm_data)
     u2.x.scatter_forward()
 
     # interpolate f which is exactly represented on the element
@@ -999,11 +1010,14 @@ def test_nonmatching_mesh_single_cell_overlap_interpolation(xtype):
     u2.interpolate(f_test2)
     u2.x.scatter_forward()
     padding = 1e-14
-    u2_2_u1_nmm_data = create_nonmatching_meshes_interpolation_data(
-        u1.function_space.mesh, u1.function_space.element, u2.function_space.mesh, padding=padding
+    cell_map1 = mesh1.topology.index_map(mesh1.topology.dim)
+    num_cells1 = cell_map1.size_local + cell_map1.num_ghosts
+    cells1 = np.arange(num_cells1, dtype=np.int32)
+    u2_2_u1_nmm_data = create_interpolation_data(
+        u1.function_space, u2.function_space, cells1, padding=padding
     )
 
-    u1.interpolate(u2, nmm_interpolation_data=u2_2_u1_nmm_data)
+    u1.interpolate_nonmatching(u2, cells1, interpolation_data=u2_2_u1_nmm_data)
     u1.x.scatter_forward()
 
     u1_exact = Function(u1.function_space, dtype=xtype)
@@ -1029,7 +1043,7 @@ def test_nonmatching_mesh_single_cell_overlap_interpolation(xtype):
 
 
 def test_submesh_interpolation():
-    """Test interpolation of a function between a submesh and its parent"""
+    """Test interpolation of a function between a sub-mesh and its parent mesh."""
     mesh = create_unit_square(MPI.COMM_WORLD, 6, 7)
 
     def left_locator(x):
@@ -1040,42 +1054,36 @@ def test_submesh_interpolation():
 
     tdim = mesh.topology.dim
     cells = locate_entities(mesh, tdim, left_locator)
-    submesh, sub_to_parent, _, _ = create_submesh(mesh, tdim, cells)
+    submesh, parent_cells, _, _ = create_submesh(mesh, tdim, cells)
 
-    V = functionspace(mesh, ("Lagrange", 2))
-    u = Function(V)
-    u.interpolate(ref_func)
+    u0 = Function(functionspace(mesh, ("Lagrange", 2)))
+    u0.interpolate(ref_func)
 
-    V_sub = functionspace(submesh, ("DG", 3))
-    u_sub = Function(V_sub)
+    V1 = functionspace(submesh, ("DG", 3))
+    u1 = Function(V1)
 
-    # Map from parent to sub mesh
-    u_sub.interpolate(u, cell_map=sub_to_parent)
+    # Interpolate u0 (defined on 'full' mesh) into u0 (defined on
+    # 'sub'0mesh)
+    u1.interpolate(u0, cells0=parent_cells, cells1=np.arange(len(parent_cells)))
 
-    u_sub_exact = Function(V_sub)
-    u_sub_exact.interpolate(ref_func)
+    u1_exact = Function(V1)
+    u1_exact.interpolate(ref_func)
     atol = 5 * np.finfo(default_scalar_type).resolution
-    np.testing.assert_allclose(u_sub_exact.x.array, u_sub.x.array, atol=atol)
+    np.testing.assert_allclose(u1_exact.x.array, u1.x.array, atol=atol)
 
     # Map from sub to parent
     W = functionspace(mesh, ("DG", 4))
     w = Function(W)
 
-    cell_imap = mesh.topology.index_map(tdim)
-    num_cells = cell_imap.size_local + cell_imap.num_ghosts
-    parent_to_sub = np.full(num_cells, -1, dtype=np.int32)
-    parent_to_sub[sub_to_parent] = np.arange(len(sub_to_parent))
-
-    # Mapping back needs to be restricted to the subset of cells in the submesh
-    w.interpolate(u_sub_exact, cells=sub_to_parent, cell_map=parent_to_sub)
-
+    # Interpolate Function defined on sub-mesh (u1_exact) to the part of
+    # a Function on the full mesh (w)
+    w.interpolate(u1_exact, cells0=np.arange(len(parent_cells)), cells1=parent_cells)
     w_exact = Function(W)
-    w_exact.interpolate(ref_func, cells=cells)
-
+    w_exact.interpolate(ref_func, cells0=cells)
     np.testing.assert_allclose(w.x.array, w_exact.x.array, atol=atol)
 
 
-def test_submesh_expression_interpolation():
+def xtest_submesh_expression_interpolation():
     """Test interpolation of an expression between a submesh and its parent"""
     mesh = create_unit_square(MPI.COMM_WORLD, 10, 8, cell_type=CellType.quadrilateral)
 
